/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of IDLBrowser.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#import <AppKit/AppKit.h>
#import <Foundation/Foundation.h>
#import "IDLViewer.h"
#import "ObjectCell.h"
#import "IDLMatrix.h"
#import "IDL.h"
#import "IORView.h"
#import "CORBAObject.h"
#import "IDLRunController.h"

@interface IDLViewer(Toolbar)
- (void) setupToolbar;
@end
@interface IDLViewer(Private)
- (BOOL)validateItemAction:(SEL)action;
@end
@interface NSObject(HeaderGen)
- (NSString*) generateHeader;
@end

@implementation IDLViewer

static NSPoint _topLeftPoint = {0,0};
static BOOL _validOffset = NO;
static NSNib* _nib = nil;

+ (void) _loadImages
{
    NSBundle* bundle = [NSBundle bundleForClass:[self class]];
    NSArray* names = [NSArray arrayWithObjects:
        @"filedrawer", @"diagram", @"propertylist", @"template", @"HeaderFile", nil];
    NSEnumerator* en = [names objectEnumerator];
    NSString* name;
    
    while(name = [en nextObject]){
        NSString* path = [bundle pathForResource:name ofType:@"tiff"];
        NSImage* image = [[NSImage alloc]initWithContentsOfFile:path];
        [image setName:name];
    }
}

+ (void) initialize
{
    if([IDLViewer class] == self){
		NSBundle* bundle = [NSBundle bundleForClass:self];
        [self _loadImages];
		_nib = [[NSNib alloc]initWithNibNamed:@"IDLExplorer.nib" bundle:bundle];
	}
}

- (void) loadInterface:sender
{
    NSArray* types = [NSArray arrayWithObjects:@"idl",@"IDL",nil];
    [parser loadInterfacesWithTypes:types window:[self window]];
}

- (void) parserDidUpdate:(NSNotification*)notification
{
	[self reload:self];
}

- (void) parse:sender
{
    [self loadInterface:self];
}

- (void) _offsetWindow
{
	if(_validOffset){
		_topLeftPoint = [window cascadeTopLeftFromPoint:_topLeftPoint];
		[window setFrameTopLeftPoint:_topLeftPoint];
	}else{
		NSRect frame = [window frame];
		_topLeftPoint.x = NSMinX(frame);
		_topLeftPoint.y = NSMaxY(frame);
		_validOffset = YES;
	}
}

- initWithParser:(IDLParser*)aParser
{
	NSArray* array;
    [super init];
    _dict = [[NSMutableDictionary alloc]init];
	_cache = [[NSMutableDictionary alloc]init];
    parser = [aParser retain];
    [[NSNotificationCenter defaultCenter]addObserver:self
        selector:@selector(parserDidUpdate:)
        name:@"ParserDidLoadInterfaces"
        object:parser];
	 if(![_nib instantiateNibWithOwner:self topLevelObjects:&array]) 
		NSLog(@"Cannot load IDLExplorer.nib");    
	_objects = [[NSMutableArray alloc]initWithArray:array];
	[_objects removeObject:[NSFontManager sharedFontManager]];
	[_objects makeObjectsPerformSelector:@selector(release)];
	
	[self _offsetWindow];
	[window makeKeyAndOrderFront:self];
	[table reloadData];
    [typesBrowser loadColumnZero];
	[table registerForDraggedTypes:[NSArray arrayWithObject:NSFilenamesPboardType]];
    [window makeFirstResponder:typesBrowser];

    return self;
}

- (void) awakeFromNib
{
    int columnCount = [[NSUserDefaults standardUserDefaults]integerForKey:@"ColumnCount"];
    [typesBrowser setTakesTitleFromPreviousColumn:NO];
    [typesBrowser setLastColumn:-1];
    [typesBrowser setMatrixClass:[IDLMatrix class]];
    [typesBrowser setCellPrototype:[[ObjectCell alloc]initTextCell:@"type"]];
    [[typesBrowser cellPrototype]setFont:[NSFont userFontOfSize:12.0]];

    [typesBrowser setAcceptsArrowKeys:YES];
    [typesBrowser setSendsActionOnArrowKeys:YES];
    [typesBrowser sendAction];
    [typesBrowser setDoubleAction:@selector(openIDLRunner:)];
    [typesBrowser setMaxVisibleColumns:columnCount];
	[typesBrowser setReusesColumns:NO];
    [objectView setDelegate:self];
    
	[table setColumnAutoresizingStyle:NSTableViewLastColumnOnlyAutoresizingStyle];
	[table setAutoresizesOutlineColumn:NO];
    
    [self setupToolbar];
}

- (void) dealloc
{
    [_dict release];
	[_cache release];
    [parser release];
    [super dealloc];
}

- window
{
	return window;
}

- (void) orderFront:sender
{
	[window makeKeyAndOrderFront:self];
}

- (void) saveDocument:sender
{
	NSString* path = [parser filename];
	if(path)
		[parser saveParserTo:path];
	else
		[self saveDocumentAs:self];
}

- (void) saveDocumentAs:sender
{
	NSSavePanel* savePanel = [NSSavePanel savePanel];
	[savePanel setRequiredFileType:@"idlcnf"];
	[savePanel beginSheetForDirectory:nil
			file:nil
			modalForWindow:window
			modalDelegate:self
			didEndSelector:@selector(savePanelDidEnd:returnCode:contextInfo:)
			contextInfo:nil];
		
    [window setDocumentEdited:NO];
}

- (void) savePanelDidEnd:(NSSavePanel*)sheet returnCode:(int)rc contextInfo:(void*)info
{
	if(rc == NSOKButton)
		[parser saveFilesTo:[sheet filename]];
}

- (void) reload:sender
{
    int n = [typesBrowser lastColumn];
    int j;
    for(j=0;j<=n;j++)
        [typesBrowser reloadColumn:j];
	[_cache removeAllObjects];
	[table reloadData];
}

- (void) _selectFileInBrowserForCell:cell
{
    id obj = [cell representedObject];
    NSString* path = [parser sourceFileForObject:obj];
	NSArray* array = path?[NSArray arrayWithObject:path] : [NSArray array];
	[fileController setSelectedObjects:array];
}

- (void) click:sender
{
    id cell = [typesBrowser selectedCell];
    if(cell){
        NSString* info = [(IDLDefinition*)[cell representedObject] stringValue];
        [typeField setStringValue:info?info:@""];
        if([drawer state] == NSDrawerOpenState)
            [self _selectFileInBrowserForCell:cell];
        if([[cell representedObject]isKindOfClass:[IDLInterface class]])
            [objectView setConnection:[[cell representedObject]interfaceConnection]];
        else{
            int col = [typesBrowser selectedColumn];
            while(col-- > 0){
                cell = [[typesBrowser matrixInColumn:col]selectedCell];
                if([[cell representedObject]isKindOfClass:[IDLInterface class]]){
                    [objectView setConnection:[[cell representedObject]interfaceConnection]];
                    break;
                }
            }
            if(col < 0)
                [objectView setConnection:nil];
        }
    }else{
        [typeField setStringValue:@""];
        [objectView setConnection:nil];
    }
    [messageField setStringValue:@""];
}

- (void)browser:(NSBrowser *)sender createRowsForColumn:(int)col inMatrix:(NSMatrix *)matrix 
{
	if(col == 0){
		int row = [table selectedRow];
		id item = (row >= 0)? [table itemAtRow:row] : nil;
		[item fillMatrix:matrix];
	}else{
		id selCell = [[sender matrixInColumn:col-1]selectedCell];
		id obj = [selCell representedObject];
		[obj fillMatrix:matrix];
	}
}

- (NSString *)browser:sender titleOfColumn:(int)column
{
    if(column == 0){
		int row = [table selectedRow];
		id item = (row >= 0)?[table itemAtRow:row] : nil;
		if(item){
			NSString* name = [item name];
			if(item == [IDLDefinitionScope rootScope])
				name = @"SystemRoot";
			return name? name : @"RootModule";
		}else
			return @"";
    }else{
        id cell = [[typesBrowser matrixInColumn:column-1]selectedCell];
		return [(NSCell*)cell stringValue];
    }
}

- selectedObject
{
    return [[typesBrowser selectedCell]representedObject];
}

- (void) windowWillClose:(NSNotification*)notifucation
{
    [[NSNotificationCenter defaultCenter]removeObserver:self];
	[_objects release];
	[self release];
}

- (void) openHeader:sender
{
    id obj = [self selectedObject];
    int ok = NO;
    if([obj respondsToSelector:@selector(generateHeader)]){
        NSString* string = [obj generateHeader];
        if(string){
            NSString* name = [obj name];
            NSString* temp = NSTemporaryDirectory();
            NSString* path = [temp stringByAppendingPathComponent:NSUserName()];
            [[NSFileManager defaultManager]createDirectoryAtPath:path withIntermediateDirectories:YES attributes:nil error:NULL];
            path = [path stringByAppendingPathComponent:
                    [NSString stringWithFormat:@"%@.h",name]];
			if([string respondsToSelector:@selector(writeToFile:atomically:usingEncoding:)])
				ok = [string writeToFile:path atomically:YES encoding:NSUTF8StringEncoding error:NULL];
			else
				ok = [string writeToFile:path atomically:YES encoding:NSUTF8StringEncoding error:NULL];
			if(ok)
                ok = [[NSWorkspace sharedWorkspace]openFile:path];
            else
                NSBeginAlertSheet(nil,nil,nil,nil,
                    window,self, NULL, NULL,  NULL,
                    @"Could not save header file to '%@'", path);
        }else
            NSBeep();
    }
}

- (void) getTemplate:sender
{
    IDLDefinition* obj = [self selectedObject];
    BOOL ok = NO;
    if([obj isKindOfClass:[IDLFunction class]] ||
            [obj isKindOfClass:[IDLAttribute class]])
        ok = [NSApp tryToPerform:@selector(openRunWindow:) with:obj];
    else
    if([obj isKindOfClass:[IDLStruct class]])
        ok = [NSApp tryToPerform:@selector(openStructWindow:) with:obj];
    if(ok)
        [messageField setStringValue:@""];
    else
        [messageField setStringValue:@"Cannot open template for this object"];
}

- (void) openIDLRunner:sender
{
    IDLDefinition* type = [self selectedObject];
	IDLDefinition* refType = [type referenceType];
	if([refType respondsToSelector:@selector(createIDLItemWithName:)]){
		IDLRunController* runner = [[IDLRunController alloc]initWithDefinition:refType];
		[runner setRootScope:[parser rootScope]];
		[[runner window]makeKeyAndOrderFront:self];
	}else
		NSBeep();
}

- (void) disconnectInterface:sender
{
	id obj = [self selectedObject];
	if([obj isKindOfClass:[IDLInterface class]]){
		[obj setInterfaceConnection:nil];
		[self click:self];
	}
}

- (BOOL)validateItemAction:(SEL)action
{
    if(action == @selector(parse:) || action == @selector(saveDocument:) || action == @selector(saveDocumentAs:))
        return ![parser isRoot];
    else
    if(action == @selector(getTemplate:)){
        id obj = [self selectedObject];
        if([obj isKindOfClass:[IDLSequence class]])
            return [[(IDLSequence*)obj sequenceType]isKindOfClass:[IDLStruct class]];
        else
            return [obj isKindOfClass:[IDLFunction class]] || 
                    [obj isKindOfClass:[IDLStruct class]] ||
                    [obj isKindOfClass:[IDLAttribute class]];
    }else
	if(action == @selector(openIDLRunner:)){
		IDLDefinition* type = [self selectedObject];
		IDLDefinition* refType = [type referenceType];
		return [refType respondsToSelector:@selector(createIDLItemWithName:)];
	}else
    if(action == @selector(openHeader:)){
        id obj = [self selectedObject];
        return [obj respondsToSelector:@selector(generateHeader)];
    }else
    if(action == @selector(disconnectInterface:)){
        id obj = [self selectedObject];
        if([obj isKindOfClass:[IDLInterface class]])
            return ([obj interfaceConnection] != nil);
        else
            return NO;
    }else
    if(action == @selector(narrowReturnType:)){
        IDLDefinition* obj = [self selectedObject];
        if([obj isKindOfClass:[IDLFunction class]]){
            if([[(IDLFunction*)obj returnType]isKindOfClass:[IDLInterface class]])
                return YES;
        }
        return NO;
    }else
        return YES;
}

- (BOOL)validateMenuItem:(NSMenuItem *)menuItem
{
    SEL action = [menuItem action];
    return [self validateItemAction:action];
}

- (BOOL) valueWillChange:sender
{
	if([[self selectedObject] isKindOfClass:[IDLInterface class]])
		return YES;
	else
		return NO;
}

- valueDidChange:sender
{
	if([[self selectedObject] isKindOfClass:[IDLInterface class]])
		[[self selectedObject]setInterfaceConnection:[sender object]];
	return self;
}

- (void) openIDLFiles:(NSArray*)files
{
	NSString* app = [[NSUserDefaults standardUserDefaults]stringForKey:@"IDLEditApp"];
	NSEnumerator* en = [files objectEnumerator];
	NSString* path;
	while(path = [en nextObject])
		[[NSWorkspace sharedWorkspace]openFile:path withApplication:app];
}

- (void)openPanelDidEnd:(NSOpenPanel *)panel 
        returnCode:(int)returnCode contextInfo:(void  *)contextInfo
{
    if(returnCode == NSOKButton){
        NSArray* files = [panel filenames];
        NSEnumerator* en = [files objectEnumerator];
        NSString* file;
        while(file = [en nextObject])
            [parser addSearchPath:file];
    }
}

- (void) addToIDLSearchPath:sender
{
    NSOpenPanel* openPanel = [NSOpenPanel openPanel];
    [openPanel setTitle:@"Add IDL Search Path"];
    [openPanel setCanChooseDirectories:YES];
    [openPanel setCanChooseFiles:NO];
    
    [openPanel beginSheetForDirectory:nil file:nil types:nil
        modalForWindow:[self window] modalDelegate:self 
        didEndSelector:@selector(openPanelDidEnd:returnCode:contextInfo:) contextInfo:NULL];
}

- (void) drawerDidOpen:(NSNotification*)notification
{
}

- (void) drawerDidClose:(NSNotification*)notification
{
}

- (void) closeSheetWillEnd:(NSWindow*)sender returnCode:(int)rc info:(void*)info
{
    switch(rc){
        case NSAlertDefaultReturn:
            [self saveDocument:self];
            break;
        case NSAlertOtherReturn:
            [window setDocumentEdited:NO];
            break;
    }
}

- (void) closeSheetDidEnd:(NSWindow*)sender returnCode:(int)rc info:(void*)info
{
    switch(rc){
        case NSAlertDefaultReturn:
        case NSAlertOtherReturn:
            [window performClose:self];
            break;
    }
}

- (BOOL) windowShouldClose:sender
{
    if([window isDocumentEdited]){
        NSBeginAlertSheet(@"Do you  want to save the model?",
            @"Save",  @"Cancel",  @"Don't save",
            window,
            self,
            @selector(closeSheetWillEnd:returnCode:info:),
            @selector(closeSheetDidEnd:returnCode:info:),
            NULL,
            @"Your changes to the idl model will be lost if you don't save them.");
        return NO;
    }else
        return YES;
}

- (void) setScope
{
    [[parser rootScope]set];
}

- (void) windowDidBecomeMain:(NSNotification*)notification
{
    [[parser rootScope]set]; // Set currentScope
    [[NSNotificationCenter defaultCenter]addObserver:self
        selector:@selector(notifyParseError:)
        name:@"IDLParseError"
        object:nil];
}

- (void) windowDidResignMain:(NSNotification*)notification
{
    [[NSNotificationCenter defaultCenter]removeObserver:self
        name:@"IDLParseError"
        object:nil];
}

- (void) _parseFiles:(NSArray*)files
{
    NSEnumerator* en = [files objectEnumerator];
    NSString* file;
    while(file = [en nextObject]){
        BOOL isDirectory = NO;
        if([[NSFileManager defaultManager]fileExistsAtPath:file isDirectory:&isDirectory]){
            if(isDirectory){
                NSArray* dir = [[NSFileManager defaultManager]contentsOfDirectoryAtPath:file error:NULL];
                [self _parseFiles:dir];
            }else
            if([[file pathExtension]isEqualToString:@"idl"]){
                [parser loadFiles:[NSArray arrayWithObject:file]];
            }
        }
    }
}

- (BOOL) view:(NSView*)view acceptsDragging:(id<NSDraggingInfo>)sender
{
    NSPasteboard* pboard = [sender draggingPasteboard];
    if([[pboard types]containsObject:NSFilenamesPboardType]){
        NSString* fileString = [pboard stringForType:NSFilenamesPboardType];
        NSArray* files = [fileString propertyList];
        BOOL ok = YES;
        @try{
			if([parser isRoot])
				ok = NO;
			else
				[self _parseFiles:files];
        }@catch(NSException *localException){
			NSBeginAlertSheet([localException name], nil,nil,nil,
				window, self, NULL,NULL,
				NULL,
				[localException reason]);
			ok = NO;
        }
		[_cache removeAllObjects]; // reset cache
		[table reloadData];
        return ok;
    }else
        return NO;
}

static NSString* pathForTypeId(NSString* typeId)
{
    if([typeId hasPrefix:@"IDL:"]){
        NSRange r = {4, [typeId length] - 4};
        NSRange r1 = [typeId rangeOfString:@":" options:0 range:r];
        if(r1.length > 0)
            r.length = r1.location - r.location;
        typeId = [typeId substringWithRange:r];
        r1 = [typeId rangeOfString:@"/"];
        if(r1.length){
			// Check if prefix is present. It has '.' lile in omg.org
            r = [typeId rangeOfString:@"." options:0 range:NSMakeRange(1,r1.location-1)];
            if(r.length > 0)
                return [typeId substringFromIndex:r1.location+1]; // Do not include '/'
        }
        return typeId;
    }else
    if([typeId hasPrefix:@"RMI:"]){
        NSRange r = {4, [typeId length] - 4};
        NSRange r1 = [typeId rangeOfString:@":" options:0 range:r];
        NSMutableString* str;
        if(r1.length > 0)
            r.length = r1.location - r.location;
        str = [NSMutableString stringWithString:[typeId substringWithRange:r]];
        [str replaceOccurrencesOfString:@"." withString:@"/" options:0 range:NSMakeRange(0,[str length])];
        [str insertString:@"/" atIndex:0];
        return str;
    }else
        return nil;
}

- (BOOL) acceptConnections:(NSArray*)array;
{
    NSEnumerator* en = [array objectEnumerator];
    id obj;
    IDLInterface* ifObj = nil;
    while(obj = [en nextObject]){
        ifObj = [obj interface];
        // Interface is nil if dragged from Naming Svc 
        if(!ifObj){
            [obj narrowInterface];
            ifObj = [obj interface];
        }
        [ifObj setInterfaceConnection:obj];
    }
    if(ifObj){
        NSString* path = pathForTypeId([ifObj typeId]);
        if(path){
			// Skip leading '/'
			NSRange r = [path rangeOfString:@"/" options:0 range:NSMakeRange(0, [path length])];
			if(r.length > 0){
				NSString* moduleName = [path substringWithRange:NSMakeRange(0, r.location)];
				NSArray* modules = [self modules];
				int row = [modules count];
				while(row-- > 0)
					if([[[modules objectAtIndex:row]name]isEqualToString:moduleName])
						break;
				if(row >= 0){
					[table selectRowIndexes:[NSIndexSet indexSetWithIndex:row] byExtendingSelection:NO];
					path = [path substringFromIndex:r.location+r.length];
				}
			}
            if([typesBrowser setPath:path])
                [typesBrowser sendAction];
		}
    }
    
    return YES;
}

#pragma mark ++++ outline view datasource and delegate ++++

- (NSArray*) modulesForItem:item
{
	NSArray* array = [_cache objectForKey:[NSNumber numberWithUnsignedInt:(unsigned)item]];
	if(!array)
		[_cache setObject:array = [item modules] forKey:[NSNumber numberWithUnsignedInt:(unsigned)item]];
	return array;
}

- (NSArray*)modules
{
	NSMutableArray* array = [NSMutableArray arrayWithObject:[[IDLParser rootParser]rootScope]];
	[array addObject:[parser rootScope]];
	[array addObjectsFromArray:[self modulesForItem:[parser rootScope]]];
	return array;
}

- (int)outlineView:(NSOutlineView *)aTable numberOfChildrenOfItem:item
{
	if(item){
		return [[self modulesForItem:item]count];
	}else{
		return [[self modulesForItem:self]count];
	}
}

- (BOOL)outlineView:(NSOutlineView *)aTable isItemExpandable:item
{
	if(item && item != [parser rootScope])
		return [[self modulesForItem:item]count] > 0;
	return NO;
}

- outlineView:(NSOutlineView *)table child:(int)index ofItem:item
{
	if(item)
		return [[self modulesForItem:item]objectAtIndex:index];
	else
		return [[self modulesForItem:self]objectAtIndex:index];
}

- outlineView:(NSOutlineView *)table objectValueForTableColumn:(NSTableColumn *)tableColumn byItem:(id)item
{
	NSTextAttachment* attachment = [[[NSTextAttachment alloc]initWithFileWrapper:nil]autorelease];
	NSTextAttachmentCell* cell = [[[NSTextAttachmentCell alloc]initImageCell:[item cellImage]]autorelease];
	NSFont* font = [[tableColumn dataCell]font];
	[attachment setAttachmentCell:cell];
	unichar c = NSAttachmentCharacter;
	NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
		attachment, NSAttachmentAttributeName,
		[NSNumber numberWithFloat:-2.0], NSBaselineOffsetAttributeName, nil];
	NSAttributedString* imageString = [[[NSAttributedString alloc]initWithString:[NSString stringWithCharacters:&c length:1]
				attributes:dict]autorelease];
	NSString* name = [item name];
	if(item == [IDLDefinitionScope rootScope])
		name = @"SystemRoot";
	NSMutableAttributedString* string = [[[NSMutableAttributedString alloc]initWithString:name?name:@"RootModule" 
		attributes:[NSDictionary dictionaryWithObject:font forKey:NSFontAttributeName]]autorelease];
	[string insertAttributedString:imageString atIndex:0];
	return string;
}

- (void)outlineViewSelectionDidChange:(NSNotification *)notification
{
	[typesBrowser loadColumnZero];
	[self click:self];
}

- (void)outlineView:(NSOutlineView *)aTable willDisplayCell:(id)cell forTableColumn:(NSTableColumn *)tableColumn item:(id)item
{
}

- (BOOL)outlineView:(NSOutlineView *)outlineView writeItems:(NSArray *)items toPasteboard:(NSPasteboard *)pboard
{
	return NO;
}

- (BOOL)outlineView:(NSOutlineView*)aTable acceptDrop:(id <NSDraggingInfo>)info 
		item:item childIndex:(int)index
{
	if(![parser isRoot]){
		NSPasteboard* pboard = [info draggingPasteboard];
		if([[pboard types]containsObject:NSFilenamesPboardType]){
			NSString* fileString = [pboard stringForType:NSFilenamesPboardType];
			NSArray* files = [fileString propertyList];
			BOOL ok = YES;
			@try{
				[self _parseFiles:files];
			}@catch(NSException *localException){
				NSBeginAlertSheet([localException name], nil,nil,nil,
					window, self, NULL,NULL,
					NULL,
					[localException reason]);
				ok = NO;
			}
			[_cache removeAllObjects]; // reset cache
			[table reloadData];
			return ok;
		}
	}
	return NO;
}

- (NSDragOperation)outlineView:(NSOutlineView *)aTable validateDrop:(id <NSDraggingInfo>)info 
		proposedItem:item proposedChildIndex:(int)index
{
	if(![parser isRoot]){
		NSPasteboard* pboard = [info draggingPasteboard];
		if([[pboard types]containsObject:NSFilenamesPboardType])
			return NSDragOperationCopy;
	}
	return NSDragOperationNone;
}

- (void) printDocument:sender
{
    [window print:sender];
}

@end

@implementation IDLViewer(Toolbar)

static NSString* IDLViewerToolbar = @"IDLViewerToolbar";
static NSString* FileDrawerItem = @"FileDrawerItem";
static NSString* TemplateItem = @"TemplateItem";
static NSString* HeaderItem = @"HeaderItem";
static NSString* ObjectViewItem = @"ObjectViewItem";

- (void) setupToolbar 
{
    NSToolbar *toolbar = [[[NSToolbar alloc]initWithIdentifier: IDLViewerToolbar]autorelease];
    [toolbar setAllowsUserCustomization: YES];
    [toolbar setAutosavesConfiguration: YES];
    [toolbar setDisplayMode: NSToolbarDisplayModeIconOnly];
    
    [toolbar setDelegate: self];
    [window setToolbar: toolbar];
}

- (NSToolbarItem *)toolbar:(NSToolbar*)toolbar 
        itemForItemIdentifier:(NSString*)itemIdent
        willBeInsertedIntoToolbar:(BOOL) willBeInserted
{
    SEL selector = NSSelectorFromString([NSString stringWithFormat:@"_get%@:",itemIdent]);
    if(selector && [self respondsToSelector:selector]){
        NSToolbarItem* item = [self performSelector:selector withObject:itemIdent];
        
        return item;
    }else
        return nil;

}

- (NSArray *)toolbarDefaultItemIdentifiers:(NSToolbar*)toolbar
{
    return [NSArray arrayWithObjects:FileDrawerItem,
        NSToolbarSeparatorItemIdentifier,
        TemplateItem, HeaderItem,
        NSToolbarSeparatorItemIdentifier,
        NSToolbarFlexibleSpaceItemIdentifier,
        NSToolbarSeparatorItemIdentifier,
        ObjectViewItem, NSToolbarSpaceItemIdentifier, nil];
        
}

- (NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar*)toolbar
{
    return [NSArray arrayWithObjects:FileDrawerItem, TemplateItem,
        HeaderItem, ObjectViewItem,
        NSToolbarSeparatorItemIdentifier,
        NSToolbarFlexibleSpaceItemIdentifier,
        NSToolbarSpaceItemIdentifier,
        NSToolbarPrintItemIdentifier,
        NSToolbarCustomizeToolbarItemIdentifier,
        nil];
}

- (void)toolbarWillAddItem:(NSNotification*)notification
{
    NSToolbarItem* item = [[notification userInfo]objectForKey:@"item"];
    if([[item itemIdentifier] isEqualToString:ObjectViewItem]){
        IORView* view = (IORView*)[item view];
        [view setDelegate:self];
    }
}

- (void)toolbarDidRemoveItem:(NSNotification*)notification
{

}

- (BOOL) validateToolbarItem:(NSToolbarItem *)toolbarItem
{
    SEL action = [toolbarItem action];
    return [self validateItemAction:action];
}

- (NSToolbarItem *)_getFileDrawerItem:(NSString*)itemIdent
{
    NSToolbarItem *item = [[[NSToolbarItem alloc]initWithItemIdentifier:itemIdent]
                                    autorelease];
    [item setLabel:@"Files"];
    [item setPaletteLabel: @"Files"];
    [item setToolTip: @"Shows IDL files in a drawer"];
    [item setImage: [NSImage imageNamed: @"filedrawer"]];
    [item setTarget:drawer];
    [item setAction:@selector(toggle:)];
    return item;
}

- (NSToolbarItem *)_getTemplateItem:(NSString*)itemIdent
{
    NSToolbarItem *item = [[[NSToolbarItem alloc]initWithItemIdentifier:itemIdent]
                                    autorelease];
    [item setLabel:@"Runner"];
    [item setPaletteLabel: @"Runner"];
    [item setToolTip: @"Open Runner window"];
    [item setImage: [NSImage imageNamed: @"template"]];
    [item setTarget:self];
    [item setAction:@selector(openIDLRunner:)];
    return item;
}

- (NSToolbarItem*)_getHeaderItem:(NSString*)itemIdent
{
    NSToolbarItem *item = [[[NSToolbarItem alloc]initWithItemIdentifier:itemIdent]
                                    autorelease];
    [item setLabel:@"Header"];
    [item setPaletteLabel: @"Header"];
    [item setToolTip: @"Generate Header file"];
    [item setImage: [NSImage imageNamed: @"HeaderFile"]];
    [item setTarget:self];
    [item setAction:@selector(openHeader:)];
    return item;
}

- (NSToolbarItem*)_getObjectViewItem:(NSString*)itemIdent
{
    NSToolbarItem *item = [[[NSToolbarItem alloc]initWithItemIdentifier:itemIdent]
                                    autorelease];
    [item setLabel:@"IOR"];
    [item setPaletteLabel: @"IOR view"];
//    [item setToolTip: @"Drag IOR to connect"];
    [item setView:objectView];
    [item setMinSize:NSMakeSize(32,24)];
    [item setMaxSize:NSMakeSize(48,32)];
    return item;
}

@end

@implementation IDLViewer(Logging)

- (void) notifyParseError:(NSNotification*)notification
{
    NSString* message = [[notification userInfo]objectForKey:@"message"];
    NSBeginAlertSheet(nil,nil,nil,nil,
        window,nil,NULL,NULL,NULL,
        message);
}

- (void)outlineViewItemDidExpand:(NSNotification *)notification
{
}

- (void)outlineViewItemDidCollapse:(NSNotification *)notification
{
}

@end
