/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import <Foundation/Foundation.h>
#import "StockServer.h"
#import "QuoteItem.h"
#import "StockAdminServer.h"
#import "NamingContextProto.h"
#import "StockWatchProto.h"

@implementation StockServer

- (IDLInterface*) interface
{
    return [[ORB defaultORB]interfaceNamed:@"StockExample/StockMarket"];
}

- initWithArray:(NSArray*)quoteArray
{
    NSEnumerator* en = [quoteArray objectEnumerator];
    NSDictionary* dict;
    self = [super init];
    _quotes = [[NSMutableDictionary alloc]init];
    _clients = [[NSMutableArray alloc]init];
    _name = @"NASDAQ";
    
    while(dict = [en nextObject]){
        QuoteItem* item = [[[QuoteItem alloc]initWithDictionary:dict]autorelease];
        [_quotes setObject:item forKey:[item valueForKey:@"symbol"]];
    }
    [[NSNotificationCenter defaultCenter]addObserver:self selector:@selector(invalidateClient:)
        name:ORBInvalidateObject object:nil];

    return self;
}

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter]removeObserver:self];
    [_admin release];
    [_clients release];
    [_name release];
    [_quotes release];
    [super dealloc];
}

- (NSMutableDictionary*) locateClient:client
{
    NSEnumerator* en = [_clients objectEnumerator];
    NSMutableDictionary* dict;
    while(dict = [en nextObject]){
        if([dict valueForKey:@"client"] == client)
            return dict;
    }
    return nil;
}

- (void) _invalidateClient:client
{
    id dict = [self locateClient:client];
    if(dict)
        [_clients removeObject:dict];
}

- (void) invalidateClient:(NSNotification*)note
{
    id client = [note object];
    [self _invalidateClient:client];
}

- (NSMutableDictionary*) addClient:client
{
    NSMutableArray* items = [NSMutableArray array];
    NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
        client, @"client",
        items, @"items",nil];
    [_clients addObject:dict];
    return dict;
}

- getStockAdmin
{
    if(!_admin)
        _admin = [[StockAdminServer alloc]initWithStock:self];
    return _admin;
}

- (void) updateQuotes
{
    NSEnumerator* en = [_quotes objectEnumerator];
    QuoteItem* item;
    while(item = [en nextObject])
        [item update];
}

- (void) addQuote:quote
{
    NSString* symbol = [quote valueForKey:@"symbol"];
    if(symbol)
        [_quotes setObject:quote forKey:symbol];
}

#pragma mark --- StockMarket IDL interface ---

- (NSString*) _getName
{
    return _name;
}

- (double) getQuoteValue:(NSString*)symbol
{
    if(symbol){
        QuoteItem* item = [_quotes objectForKey:symbol];
        if(item){
            [item update];
            return [item quote];
        }
    }
    [[CORBAException exceptionWithName:@"UnknownSymbol" interface:[self interface]]raise];
    return 0;
}

- (QuoteItem*) subscribe:client symbol:(NSString*)symbol
{
    if(symbol){
        QuoteItem* item = [_quotes objectForKey:symbol];
        if(item){
            NSDictionary* clientDict = [self locateClient:client];
            NSMutableArray* items;
            if(!clientDict)
                clientDict = [self addClient:client];
            items = [clientDict objectForKey:@"items"];
            if(![items containsObject:item])
                [items addObject:item];
            return item;
        }
    }
    [[CORBAException exceptionWithName:@"UnknownSymbol" interface:[self interface]]raise];
    return nil;
}


- (void) unsubscribe:client symbol:(NSString*)symbol
{
    if(symbol){
        QuoteItem* item = [_quotes objectForKey:symbol];
        if(item){
            NSDictionary* clientDict = [self locateClient:client];
            NSMutableArray* items = [clientDict objectForKey:@"items"];
            [items removeObject:item];
            return;
        }
    }
}

- (void) start:client
{
    NSMutableDictionary* dict = [self locateClient:client];
    [dict setObject:[NSDate date] forKey:@"started"];
}

- (void) stop:client
{
    NSMutableDictionary* dict = [self locateClient:client];
    [dict removeObjectForKey:@"started"];
}

- (void) notifyClients:(NSTimer*)timer
{
    NSEnumerator* en = [_clients objectEnumerator];
    NSDictionary* dict;
    [self updateQuotes];
    while(dict = [en nextObject])
        if([dict objectForKey:@"started"]){
            id client = [dict objectForKey:@"client"];
            @try{
				[client notify:[dict objectForKey:@"items"]];
            }@catch(NSException *localException){
				if([[localException name]isEqualToString:CORBASystemException]){
					NSRange r = [[localException reason]rangeOfString:@"INV_OBJREF"];
					if(r.length > 0)
						[self _invalidateClient:client];
				}
				if([[localException name]isEqualToString:CORBACommunicationException])
					[self _invalidateClient:client];
				NSLog(@"%@: %@", [localException name], [localException reason]);
            }
        }
}

+ (void) run
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc]init];
    NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
    NSString* path = [defaults stringForKey:@"quotes"];
    NSString* iorPath = [defaults stringForKey:@"ior"];
    NSArray* array = nil;
    StockServer* server;
    NSString* ior;
    id nameService = [ORB resolveInitialReferences:@"NameService"];

    if(path){
        NSString* str = [NSString stringWithContentsOfFile:path encoding:NSUTF8StringEncoding error:NULL];
        array = [str propertyList];
    }
    @try{
		server = [[[StockServer alloc]initWithArray:array]autorelease];
		[server activate];
		ior = [[server objectReference]toString];
		if(iorPath){
			[ior writeToFile:iorPath atomically:NO encoding:NSUTF8StringEncoding error:NULL];
			NSLog(@"StockServer saved IOR in the file '%@'\nYou can open this IOR with StockApp to watch quotes", iorPath);
		}
		if(nameService)
			[nameService rebind:[IDLNameComponent nameComponentsForPath:@"StockServer"] object:server];
		[NSTimer scheduledTimerWithTimeInterval:3.0 
			target:server selector:@selector(notifyClients:) userInfo:nil repeats:YES];
    }@catch(NSException *localException){
		NSLog(@"%@: %@", [localException name], [localException reason]);
    }
    [pool release];
    
    pool = [[NSAutoreleasePool alloc]init];
    [ORB run];
    [pool release];
}

@end

int main(int argc, char* argv[])
{
    [StockServer run];
    
    return 0;
}

