/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "StockWatch.h"
#import <ADORB/ADORB.h>
#import <AppKit/AppKit.h>
#import "QuoteItem.h"
#import "StockController.h"
#import "StockMarket.h"

@interface NSObject(Controller)
- (CORBAObject*) server;
@end

@implementation StockWatch

+ (void) initialize
{
//	[self setKeys:[NSArray arrayWithObject:@"watching"]triggerChangeNotificationsForDependentKey:@"watchTitle"];
}

+ (NSSet*)keyPathsForValuesAffectingWatchTitle
{
	return [NSSet setWithObject:@"watching"];
}

- (IDLInterface*) interface
{
    return [[ORB defaultORB]interfaceNamed:@"StockExample/StockWatch"];
}

- initWithController:controller
{
    self = [super init];
	_controller = controller;
    if([NSBundle loadNibNamed:@"StockWatch" owner:self]){
        _quoteArray = [[NSMutableArray alloc]init];
        [[NSNotificationCenter defaultCenter] addObserver:self 
                selector:@selector(updateQuotes:)
                name:@"QuotesDidUpdate"
                object:self];
		// we want to be notified when the controller is re-connected to server
		[_controller addObserver:self forKeyPath:@"server" options:0 context:NULL];
        return self;
    }else{
        [self release];
        return nil;
    }
}

- (void) dealloc
{
	[_controller removeObserver:self forKeyPath:@"server"];
    // Remove ourselves from all the connections
    [self deactivate];
    [[NSNotificationCenter defaultCenter]removeObserver:self];
    [_quoteArray release];
    [super dealloc];
}

- (void) windowWillClose:(NSNotification*)note
{
	[_controller removeWatch:self];
}

- (NSArray*) quoteArray
{
	return _quoteArray;
}

- (void) addSymbol
{
    @try{
		NSString* symbol = [[symbolField stringValue]uppercaseString];
		QuoteItem* quote = [(id)[_controller server] subscribe:self symbol:symbol];
		if(![_quoteArray containsObject:quote]){
			[self willChange:NSKeyValueChangeInsertion valuesAtIndexes:[NSIndexSet indexSetWithIndex:[_quoteArray count]] forKey:@"quoteArray"];
			[_quoteArray addObject:quote];
			[self didChange:NSKeyValueChangeInsertion valuesAtIndexes:[NSIndexSet indexSetWithIndex:[_quoteArray count]] forKey:@"quoteArray"];
		}
		[symbolField setStringValue:symbol];
    }@catch(NSException *localException){
		NSRunAlertPanel(@"Exception", @"%@\n%@",nil,nil,nil,[localException name], [localException reason]);
    }
}

- (void) addSymbol:sender
{
	[NSApp beginSheet:addPanel modalForWindow:window modalDelegate:self didEndSelector:NULL contextInfo:NULL];
}

- (void) set:sender
{
	[NSApp endSheet:addPanel];
	[addPanel orderOut:self];
	if([sender tag])
		[self addSymbol];
}

- (void) removeQuote:(QuoteItem*)quote
{
	int row = [_quoteArray indexOfObject:[quote valueForKey:@"self"]];
	if(row != NSNotFound){
		@try{
			[(id)[_controller server] unsubscribe:self symbol:[symbolField stringValue]];
		}@catch(NSException *localException){
			NSRunAlertPanel(@"Exception", @"%@\n%@",nil,nil,nil,[localException name], [localException reason]);
		}
		[self willChange:NSKeyValueChangeRemoval valuesAtIndexes:[NSIndexSet indexSetWithIndex:row] forKey:@"quoteArray"];
		[_quoteArray removeObjectAtIndex:row];
		[self didChange:NSKeyValueChangeRemoval valuesAtIndexes:[NSIndexSet indexSetWithIndex:row] forKey:@"quoteArray"];
	}
}

- (BOOL) isWatching
{
	return _isWatching;
}

- (void) setWatching:(BOOL)flag
{
	@try{
		if(flag){
			if(!_isWatching){
				_isWatching = YES;
				[(id)[_controller server] start:self];
			}
		}else{
			if(_isWatching){
				_isWatching = NO;
				[(id)[_controller server] stop:self];
			}
		}
	}@catch(NSException *localException){
		NSRunAlertPanel(@"Exception", @"%@\n%@",nil,nil,nil,[localException name], [localException reason]);
    }
}

- (NSString*) watchTitle
{
	return _isWatching? @"Stop" : @"Watch";
}

- (void) toggleWatch:sender
{
	[self setWatching:![self isWatching]];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
	// object is always _controller and keyPath = @"server"
	id server = [_controller server];
	if(server){ // gets connected after being disconnected
		NSEnumerator *en = [_quoteArray objectEnumerator];
		id quote;
		@try{
			// Subscribe again to our symbols
			while(quote = [en nextObject])
				[server subscribe:self symbol:[quote valueForKey:@"symbol"]];
			// Start watching if we did before
			if(_isWatching)
				[server start:self];
		}@catch(NSException *localException){
			NSLog(@"%@, %@", [localException name], [localException reason]);
		}
	}
	// Do nothing if get disconnected, just keep the state
}

- (void) notify:(NSArray*)quotes
{
    // Post notification on main thread
    [ORB postNotification:[NSNotification notificationWithName:@"QuotesDidUpdate" object:self 
		userInfo:[NSDictionary dictionaryWithObject:quotes forKey:@"quotes"]]];
}

- (void) updateQuotes:(NSNotification*)note
{
	NSArray* quotes = [[note userInfo]objectForKey:@"quotes"];
	if([quotes count] != [_quoteArray count]){
		[self willChange:NSKeyValueChangeRemoval valuesAtIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_quoteArray count])] forKey:@"quoteArray"];
		[_quoteArray removeAllObjects];
		[self didChange:NSKeyValueChangeRemoval valuesAtIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_quoteArray count])] forKey:@"quoteArray"];
		[self willChange:NSKeyValueChangeInsertion valuesAtIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [quotes count])] forKey:@"quoteArray"];
		[_quoteArray addObjectsFromArray:quotes];
		[self willChange:NSKeyValueChangeInsertion valuesAtIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [quotes count])] forKey:@"quoteArray"];
	}else{
		[self willChange:NSKeyValueChangeReplacement valuesAtIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [quotes count])] forKey:@"quoteArray"];
		[_quoteArray replaceObjectsAtIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [_quoteArray count])] withObjects:quotes];
		[self didChange:NSKeyValueChangeReplacement valuesAtIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [quotes count])] forKey:@"quoteArray"];
	}
}

- window
{
    return window;
}

@end
