/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "DemoClient.h"
#import <ADORB/ADORB.h>
#import "NamingContextProto.h"

@interface FPoint:NSObject
{
    float _x;
    float _y;
}
- initWidth:(float)x height:(float)y;
+ point:(float)x :(float)y;
@end

@implementation DemoClient

- init
{
	NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
	NSString* nsURL = [defaults stringForKey:@"ORBNamingURL"];
	if(!nsURL){
		NSString* nsPort = [defaults stringForKey:@"port"];
		if(!nsPort)
			nsPort = @"15600"; // The default orbd's ORBInitialPort is different
		// Form corbaloc URL
		nsURL = [NSString stringWithFormat:@"corbaloc:iiop:1.2@localhost:%@", nsPort];
	}
	// This is the path for our object in the naming service
	NSString* path = @"demo/DemoServer";
    self = [super init];
	_name = @"client";
	ORB* orb = [ORB defaultORB]; // force to parse IDLs
    @try{
		[CORBAObject setDefaultConnectTimeout:10.0];
		// connect to naming server
		CORBAObject* ns = [CORBAObject objectWithURL:nsURL];
		// corbaloc provides no information of what interface 'ns' implements
		// so we need to set interface explicitly
		// Get interface object for NamingContext and set it
		IDLInterface* nsInterface = [orb interfaceNamed:@"CosNaming/NamingContext"];
		[ns setInterface:nsInterface];
		[ns setConnectTimeout:10.0];
		server = [[ns resolve:[IDLNameComponent nameComponentsForPath:path]]retain];
		[server setConnectTimeout:10.0];
    }@catch(NSException *localException){
		NSLog(@"%@: %@\n", [localException name], [localException reason]);
    }
    return self;
}

- (void) dealloc
{
    [server release];
	[_name release];
    [super dealloc];
}

static NSString* interfaceName = @"demo/Client";

// This method allows us to receive callbacks from the server
- (IDLInterface*) interface
{
    IDLInterface *obj = [[ORB defaultORB]interfaceNamed:interfaceName];
	if(obj != nil)
		return obj;
	else{
		NSLog(@"*** No interface for path '%@'", interfaceName);
		return nil;
	}
}

// Method of demo/Client interface
- (NSString*) getName
{
	return [NSString stringWithFormat:@"<%@ 0x%x>", [self className], self];
}

- (void) run
{
	NSLog(@"%@: starting...\n", [self className]);
    @try{
		NSArray * path = [NSArray arrayWithObjects:[FPoint point:0.0:0.0], [FPoint point:1.0:0.0], [FPoint point:1.0:1.0], nil];
		NSArray *path1 = [server echoPath:path];
		// Check that returned NSArray contains objects of class FPoint and is equal
		NSAssert([path1 isEqual:path], @"failed");
		
		// we will receive callback from within this remote call
		NSLog(@"-> get_access_for ...");
		id obj = [server getAccessFor:self];
		NSAssert(server == [obj valueForKey:@"object"], @"failed");
		NSString* name = [obj valueForKey:@"name"];
		NSAssert([name isEqualToString:[self getName]], @"failed");
		
		IDLDefinition* refType = [[ORB defaultORB]definitionNamed:@"demo/Pair"];
		IDLObject* pair = [IDLObject objectWithIDLType:refType];
		[pair setValue:@"access" forKey:@"name"];
		[pair setValue:obj forKey:@"value"];
		
		IDLDefinition* type = [[ORB defaultORB]definitionNamed:@"demo/Node"];
		// Node defines a node in a double-linked chain
		// Create a triangle
		// note that this will create memory leaks in server
		IDLObject* node0 = [IDLObject objectWithIDLType:type];
		IDLObject* node1 = [IDLObject objectWithIDLType:type];
		IDLObject* node2 = [IDLObject objectWithIDLType:type];
		[node0 setValue:pair forKey:@"ref"];
		[node1 setValue:pair forKey:@"ref"];
		[node2 setValue:pair forKey:@"ref"];
		[node0 setValue:@"node-0" forKey:@"name"];
		[node1 setValue:@"node-1" forKey:@"name"];
		[node2 setValue:@"node-2" forKey:@"name"];
		[node0 setValue:node1 forKey:@"next"];
		[node1 setValue:node2 forKey:@"next"];
		[node2 setValue:node0 forKey:@"next"];
		[node0 setValue:node2 forKey:@"prev"];
		[node2 setValue:node1 forKey:@"prev"];
		[node1 setValue:node0 forKey:@"prev"];
		NSLog(@"-> echo_node ...");
		IDLObject* rnode = [server echoNode:node0];
		// The returned value is equivalent 
		// rnode->next->next == rnode->prev
		NSAssert([rnode valueForKey:@"prev"] == [[rnode valueForKey:@"next"]valueForKey:@"next"], @"failed");
		NSAssert([rnode valueForKey:@"ref"] == [[rnode valueForKey:@"next"] valueForKey:@"ref"], @"failed");
		
		IDLAnyObject* any = [IDLAnyObject objectWithIDLType:type];
		[any setValue:node0];
		NSLog(@"-> echo_any ...");
		IDLAnyObject* any1 = [server echoAny:any];
		rnode = [any1 value];
		NSAssert([rnode valueForKey:@"prev"] == [[rnode valueForKey:@"next"]valueForKey:@"next"], @"failed");
		NSAssert([rnode valueForKey:@"ref"] == [[rnode valueForKey:@"next"] valueForKey:@"ref"], @"failed");	
		
		NSLog(@"*** DemoClient passed ***");
    }@catch(NSException *localException){
		NSLog(@"Exception %@: %@\n", [localException name], [localException reason]);
    }
}

@end

@implementation FPoint
- initWidth:(float)x height:(float)y
{
    self = [super init];
    _x = x;
    _y = y;
    return self;
}
- init
{
    return [self initWidth:0.0 height:0.0];
}

+ point:(float)x :(float)y
{
    return [[[self alloc]initWidth:x height:y]autorelease];
}

- (NSString*)description
{
    return [NSString stringWithFormat:@"<FPoint 0x%x (%f,%f)>", self, _x,_y];
}

- (BOOL) isEqual:(FPoint*)p
{
	return [p isMemberOfClass:[self class]] && _x == p->_x && _y == p->_y;
}

@end
