/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLDefinition_Private.h"
#import "IDLValueClass.h"
#import "IDLValueType_Private.h"
#import "IDLArrayIdentifier.h"
#import "IDLArray.h"
#import "IDLValueMember.h"
#import "IDLTypedef.h"
#import "IDLInterface.h"
#import "IDLEnum.h"
#import "CORBACoding.h"
#import "IDLAbstractBase.h"

@implementation IDLValueClass

- initWithValueType:(IDLValueType*)valuetype;
{
    self = [super init];
	_valuetype = valuetype;
    return self;
}

- (void) dealloc
{
    [_definitionDict release];
    [_supportArray release];
    [_inheritance release];
    [_super release];
    [_interface release];
    [_members release];
    [super dealloc];
}

- (void) setModifier:(int)modifier
{
    _modifier = modifier;
}

- (NSMutableDictionary*)definitionDictionary
{
    if(!_definitionDict)
        _definitionDict = [[NSMutableDictionary alloc]init];
    return _definitionDict;
}

- (NSMutableArray*)support
{
    if(!_supportArray)
        _supportArray = [[NSMutableArray alloc]init];
    return _supportArray;
}

- (NSMutableArray*)inheritance
{
    if(!_inheritance)
        _inheritance = [[NSMutableArray alloc]init];
    return _inheritance;
}

- (NSMutableArray*) members
{
    if(!_members)
        _members = [[NSMutableArray alloc]init];
    return _members;
}

- (IDLDefinition*) _definitionNamed:(NSString*)name
{
    return [_definitionDict objectForKey:name];
}

- (id <IDLDefinition>) localDefinitionNamed:(NSString*)name
{
    // Look locally, in _super, _inheritance, _interface, _supportArray, _parentScope
    IDLDefinition* obj = [self _definitionNamed:name];
    NSEnumerator* en;
    id scope;
    if(obj)
        return obj;
    if(obj = (IDLDefinition*)[_super localDefinitionNamed:name])
        return obj;
    
    en = [_inheritance objectEnumerator];
    while(scope = [en nextObject])
        if(obj = (IDLDefinition*)[scope localDefinitionNamed:name])
            return obj;
    if(obj = (IDLDefinition*)[_interface localDefinitionNamed:name])
        return obj;
    
    en = [_supportArray objectEnumerator];
    while(scope = [en nextObject])
        if(obj = (IDLDefinition*)[scope localDefinitionNamed:name])
            return obj;
    return nil;
}

- (void) addDefinition:(id <IDLDefinition>)def withName:(NSString*)name
{
    if(def && name)
        [[self definitionDictionary] setObject:def forKey:name];
}

- addDefinition:(id <IDLDefinition>)def withIdentifierArray:(NSArray*)identArray
{
    NSMutableArray* array = [NSMutableArray array];
    id identifier;
    NSEnumerator* en = [identArray objectEnumerator];
    while(identifier = [en nextObject]){
        NSString* name;
        BOOL isArray = [identifier isKindOfClass:[IDLArrayIdentifier class]];
        if(isArray)
            name = [identifier name];
        else
            name = identifier;
                    
        if([self _definitionNamed:name]){
            NSLog(@"Redefining name '%@'\n",name); 
        }else{
            IDLTypedef* tdef;
            if(isArray){
                IDLArray* obj = [IDLArray arrayWithType:def identifier:identifier];
                tdef = [IDLTypedef typedefWithType:obj name:name];
            }else
                tdef = [IDLTypedef typedefWithType:(IDLDefinition*)def name:name];
            [self addDefinition:tdef withName:name];
			[array addObject:tdef];
        }
    }
    return [array count] > 0? array : nil;
}

- (void) addDefinitionsFromDictionary:(NSDictionary*)dict
{
    [[self definitionDictionary] addEntriesFromDictionary:dict];
}

- (NSDictionary*)_definitionDictionary
{
    return _definitionDict;
}

- (id <IDLDefinition>)definitionNamed:(NSString*)name
{
    id<IDLDefinition> obj = [self localDefinitionNamed:name];
    return obj? obj : [[self parentScope] definitionNamed:name];
}

- (id <IDLDefinition>) definitionForNamePath:(NSString*)path
{
    if([path length] > 0){
        NSRange r = [path rangeOfString:@"/"];
        if(r.length > 0){
            if(r.location > 0){
                IDLDefinition* obj = (IDLDefinition*)[self definitionNamed:[path substringToIndex:r.location]];
                if([obj respondsToSelector:_cmd])
                    return (IDLDefinition*)[(id<IDLDefinitionScope>)obj definitionForNamePath:[path substringFromIndex:r.location+r.length]];
                else
                    return nil;
            }else
                return (IDLDefinition*)[[self rootScope]definitionForNamePath:[path substringFromIndex:r.length]];
        }else
            return (IDLDefinition*)[self definitionNamed:path];
    }
    return nil;
}

// 'typeId' always contains an absolute path and this method is always sent to the root scope
// 'typeId' is defined ambigiously - first name in the 'path' may be prefix or module
- (id <IDLDefinition>) definitionForTypeId:(NSString*)typeId
{
    return [[self rootScope] definitionForTypeId:typeId];
}

- (NSNumber*) enumValueForName:(NSString*)n andKey:(NSString*)key 
{
    if(n && key){
        IDLEnum* def = (IDLEnum*)[self definitionForNamePath:n];
        if(def)
            return [def enumValueForKey:key];
        else
            [NSException raise:IDLInvalidArgumentException
                format:@"Cannot locate enum definition for '%@'", n];
    }else
        [NSException raise:IDLInvalidArgumentException 
            format:@"NULL argument in [%@ %@]",
            NSStringFromClass([self class]),NSStringFromSelector(_cmd)];
    
    return nil;
}

- (id <IDLDefinitionScope>) parentScope
{
    return _parentScope;
}

- (IDLDefinitionScope*) rootScope
{
    id scope = [self parentScope];
    if(scope)
        return [scope rootScope];
    else
        return nil;
}

- (void) setParentScope:(id<IDLDefinitionScope>)scope
{
    if(!_parentScope)
        _parentScope = scope;
    else
        NSLog(@"Trying to re-set parent scope for %@ to %@\n",
            [self description],
            [(NSObject*)scope description]);
}

- (NSString*)name
{
    return @"valueclass";
}

- (void) setName:(NSString*)name
{
}

- (void) setInheritance:(NSArray*)list truncatable:(BOOL)flag
{
    NSEnumerator* en = [list objectEnumerator];
    IDLValueType* value;
    while(value = [en nextObject]){
        IDLValueType* obj = (IDLValueType*)[value referenceType];
        if([obj isKindOfClass:[IDLValueType class]]){
            if([obj isAbstract])
                [[self inheritance]addObject:obj];
            else{
                if(!_super)
                    _super = [obj retain];
                else
				if(obj != _super)
					// I do not know how to recover Yacc after an exception. So I use NSLog.
					NSLog(@"CORBAParserError: valuetype is already derived from a concrete valuetype. Trying to reset one.");
            }
        }else
        if(obj)
            NSLog(@"CORBAParserError: valuetype derived from not a valuetype, %@", 
                [obj description]);
        else{
            NSLog(@"CORBAParserError: valuetype cannot resolve %@", [value description]);
            if(!_super)
                _super = [value retain];
            else
				NSLog(@"CORBAParserError: valuetype derived from more than one concrete valuetype");
        }
            
    }
    _truncatable = flag?YES:NO;
}

- (void) setSupportedInterfaces:(NSArray*)list
{
    NSEnumerator* en = [list objectEnumerator];
    IDLInterface* obj;
    while(obj = [en nextObject]){
        if([obj isKindOfClass:[IDLInterface class]]){
            if([obj isAbstract])
                [[self support]addObject:obj];
            else{
                if(!_interface)
                    _interface = [obj retain];
                else
					NSLog(@"CORBAParserError: more than one concrete interface in valuetype support clause");
            }
        }else
			NSLog(@"CORBAParserError: not an interface in valuetype support clause");
    }
}

extern NSString* currentParserFile();

- (BOOL) addMembers:(NSArray*)identArray ofType:(IDLDefinition*)type withMode:(int)accessMode
{
    BOOL ok = YES;
    id identifier;
    NSEnumerator* en = [identArray objectEnumerator];
    while(identifier = [en nextObject]){
        IDLDefinition* obj;
        NSString* name;
        BOOL isArray = [identifier isKindOfClass:[IDLArrayIdentifier class]];
        if(isArray)
            name = [identifier name];
        else
            name = identifier;
                    
        if(isArray){
            IDLArray* arrayType = [IDLArray arrayWithType:type identifier:identifier];
            [[self members] addObject:
                obj = [IDLValueMember valueMemberOfType:arrayType mode:accessMode name:name]];
        }else{
            [[self members] addObject:
                obj = [IDLValueMember valueMemberOfType:type mode:accessMode name:name]];
        }
    }
    return ok;
}

- (BOOL) isAbstract
{
    return _modifier == vt_modifier_abstract;
}

- (int) modifier
{
    return _modifier;
}
- superValue
{
    return _super;
}

- (NSString*) typeId
{
	return [_valuetype typeId];
}

- (BOOL) isDerivedFrom:(IDLValueType*)value
{
    NSEnumerator* en;
    IDLValueType* obj;
    if([_super isDerivedFrom:value])
        return YES;
    
    en = [_inheritance objectEnumerator];
    while(obj = [en nextObject])
        if([obj isDerivedFrom:value])
            return YES;
            
    return NO;
}

// Check only abstract interfaces
// Value is not derived from a concrete interface
- (BOOL) isDerivedFromInterface:(IDLInterface*)interface
{
    // We check '_interface' because valutype is NOT derived from it.
    if(interface == nil || interface == [IDLAbstractBase sharedInstance])
        return YES;
    else{
        NSEnumerator* en = [_supportArray objectEnumerator];
        IDLInterface* obj;
        IDLValueType* value;
        while(obj = [en nextObject]){
            if([obj adoptsInterface:interface])
                return YES;
        }
        
        if([_super isDerivedFromInterface:interface])
            return YES;
        
        en = [_inheritance objectEnumerator];
        while(value = [en nextObject]){
            if([value isDerivedFromInterface:interface])
                return YES;
        }
        return NO;
    }
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	if([coder allowsKeyedCoding]){
		[coder encodeObject:_definitionDict forKey:@"definitions"];
		[coder encodeConditionalObject:_parentScope forKey:@"parentScope"];
		[coder encodeObject:_interface forKey:@"interface"];
		[coder encodeObject:_super forKey:@"super"];
		[coder encodeObject:_supportArray forKey:@"support"];
		[coder encodeObject:_inheritance forKey:@"inheritance"];
		[coder encodeObject:_members forKey:@"members"];
		[coder encodeInt:_modifier forKey:@"modifier"];
		[coder encodeBool:_truncatable forKey:@"truncatable"];
	}else{
		[coder encodeObject:_definitionDict];
		[coder encodeConditionalObject:_parentScope];
		[coder encodeObject:_interface];
		[coder encodeObject:_super];
		[coder encodeObject:_supportArray];
		[coder encodeObject:_inheritance];
		[coder encodeObject:_members];
		[coder encodeValueOfObjCType:@encode(short) at:&_modifier];
		[coder encodeValueOfObjCType:@encode(BOOL) at:&_truncatable];
	}
}

- initWithCoder:(NSCoder*)coder
{
	self = [super init];
	if([coder allowsKeyedCoding]){
		_definitionDict = [[coder decodeObjectForKey:@"definitions"]retain];
		_parentScope = [[coder decodeObjectForKey:@"parentScope"]retain];
		_interface = [[coder decodeObjectForKey:@"interface"]retain];
		_super = [[coder decodeObjectForKey:@"super"]retain];
		_supportArray = [[coder decodeObjectForKey:@"support"]retain];
		_inheritance = [[coder decodeObjectForKey:@"inheritance"]retain];
		_members = [[coder decodeObjectForKey:@"members"]retain];
		_modifier = [coder decodeIntForKey:@"modifier"];
		_truncatable = [coder decodeBoolForKey:@"truncatable"];
	}else{
		_definitionDict = [[coder decodeObject]retain];
		_parentScope = [[coder decodeObject]retain];
		_interface = [[coder decodeObject]retain];
		_super = [[coder decodeObject]retain];
		_supportArray = [[coder decodeObject]retain];
		_inheritance = [[coder decodeObject]retain];
		_members = [[coder decodeObject]retain];
		[coder decodeValueOfObjCType:@encode(short) at:&_modifier];
		[coder decodeValueOfObjCType:@encode(BOOL) at:&_truncatable];
	}
	return self;
}


@end
