/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "VXConnection.h"
#import "VXMessageHandler.h"
#import "VXMessageHeader.h"
#import "VXMessage.h"
#import "VXInvocation.h"
#import "ORB_Private.h"
#import "CORBAObject_Private.h"
#import "IDLDefinitionScope.h"
#import "IDLDefinition_Private.h"
#import "IDLEncapsulation.h"
#import "IDLServiceContext.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/time.h>
#include <unistd.h>

@interface VXBuffer : NSObject
{
@public
	NSMutableData* data;
	VXMessageHeader* header0;
	VXMessageHeader* header;
	IDLOctetStream* stream;
	unsigned length;
}
@end
@interface VXWriteBuffer : NSObject
{
@public
	VXWriteBuffer* next;
	NSData* data;
	unsigned offset;
	unsigned length;
}
- initWithData:(NSData*)aData;
@end

enum VXReadState {
	VXReadHeader = 0,
	VXReadBody,
	VXReadFragmentHeader,
	VXReadFragmentLength,
	VXReadFragmentBody,
	VXReadFragmentCancel
};

static VXConnection *lookupConnectionByAddr(struct sockaddr_in* addr);
static void addConnection(VXConnection* conn);
static void setupSocket(int fd);
extern int close(int fd);
static NSString* addressHost(struct sockaddr_in* addr);
static NSNumber* addressPort(struct sockaddr_in* addr);

@implementation VXConnection

NSString* VXInvalidateConnection = @"CORBAInvalidConnection";
NSString* VXAcceptedConnection = @"CORBADidAcceptConnection";
NSString* VXTimerDidFire = @"CORBATimerDidFire";
NSString* ConnectionException = @"ConnectionException";

static NSMutableDictionary* _hostCache = nil;
static BOOL _running = NO;
#if USE_HASHTABLES
static NSHashTable *connections = NULL;
#else
static CFSetCallBacks _connectionCallBacks = {
	0, NULL, NULL, NULL, NULL, NULL
};
static CFMutableSetRef connections = NULL;
#endif
static NSMutableArray* _acceptedConnections = nil;
static int enableLogging = NO;
static NSLock* _hlock = nil;
static NSLock* _connLock = nil;
static NSLock* _runLock = nil;
static NSLock* _bindLock = nil;
static NSPipe* _pipe = nil; // to break 'select()'
static VXConnection* _listener = nil;
static NSLock* _invocationLock = nil;
static NSMutableArray* _invocationQueue = nil;
static short _listenerPort = 0;
static struct timeval _timeout = {0,0};
static struct timeval _tm = {0,0};
static int tcp_nodelay = YES;

static int numberOfThreads = 0;
static int maxNumberOfThreads = 10;
static BOOL handleInvocationsOnMainThread = NO;

static BOOL _trace = NO;

+ (NSArray*)allObjects
{
#if USE_HASHTABLES
	return NSAllHashTableObjects(connections);
#else
	return [(NSSet*)connections allObjects];
#endif
}

+ (void) enableLogging:(BOOL)flag
{
    enableLogging = flag?YES:NO;
}

+ (void) setTimeout:(NSTimeInterval)sec
{
    if(sec > 0){
        _timeout.tv_sec = floor(sec);
        _timeout.tv_usec = floor((sec - floor(sec))*1000000);
    }else
        _timeout.tv_sec = _timeout.tv_usec = 0;
}

+ (NSTimeInterval) timeout
{
    return _timeout.tv_sec + 0.000001*_timeout.tv_usec;
}

+ (struct timeval*)_timeout
{
    if(_timeout.tv_sec == 0 && _timeout.tv_usec == 0)
        return NULL;
    else{
        _tm.tv_sec = _timeout.tv_sec;
        _tm.tv_usec = _timeout.tv_usec;
        return &_tm;
    }
}

+ (unsigned) maxNumberOfInvocationThreads
{
	return maxNumberOfThreads;
}

+ (void) setMaxNumberOfInvocationThreads:(unsigned)num
{
	maxNumberOfThreads = num;
}

+ (void) initialize
{
    if([VXConnection class] == self){
        NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
        NSString* str = [defaults stringForKey:@"debug"];
        if(str && [str rangeOfString:@"l"].length > 0)
            enableLogging = YES;
        _pipe = [[NSPipe alloc]init];
        _runLock = [[NSLock allocWithZone:[self zone]]init];
        _bindLock = [[NSLock allocWithZone:[self zone]]init];
        _hlock = [[NSLock alloc]init];
        _connLock = [[NSLock alloc]init];
        _invocationLock = [[NSLock alloc]init];
        _invocationQueue = [[NSMutableArray alloc]init];
		_hostCache = [[NSMutableDictionary alloc]init];
        _trace = [defaults boolForKey:@"trace"];
    }
    
    if(!connections){
        NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
        _listenerPort = [defaults integerForKey:@"ORBListenerPort"];
#if USE_HASHTABLES
        connections = NSCreateHashTable(NSNonRetainedObjectHashCallBacks, 0);
#else
		connections = CFSetCreateMutable(NULL, 0, &_connectionCallBacks);
#endif
        _acceptedConnections = [[NSMutableArray alloc]init];
        signal(SIGPIPE,SIG_IGN);
        [[NSNotificationCenter defaultCenter]addObserver:self
                        selector:@selector(invalidateConnection:)
                        name:VXInvalidateConnection
                        object:nil];
    }
}

- init
{
    self = [super init];
    _lock = [[NSLock allocWithZone:[self zone]]init];
    _invocations = [[NSMutableDictionary allocWithZone:[self zone]]init];
	_stringEncoding = NSUTF8StringEncoding;
	_wstringEncoding = NSUnicodeStringEncoding;
	_fd = -1;
    return self;
}

- (void) dealloc
{
    [_hlock lock];
    if(!_flags.released){
#if USE_HASHTABLES
        NSHashRemove(connections,self);
#else
		CFSetRemoveValue(connections, self);
#endif
        _flags.released = YES; // Just for a case
    }
    [_hlock unlock];
    if(_fd > 0){
        close(_fd);
        if(enableLogging)
            NSLog(@"VXConnection 0x%x: closed socket %d\n", self, _fd);
    }
	if(_inetAddr)
		NSZoneFree([self zone], _inetAddr);
    [_invocations release];
    [_objects release];
	[_readBuffer release];
	[_writeBuffer release];
    [_address release];
    [_hostname release];
	[_listenPoints release];
    [_lock release];
    [super dealloc];
}

+ (void) invalidateConnection:(NSNotification*)notification
{
    id conn = [notification object];
    [_connLock lock];
    if(conn == _listener){
        _listener = nil;
        [conn release];
    }
    [_acceptedConnections removeObject:conn];
    [_connLock unlock];
}

- (void) listen
{
    if(_flags.mode == VXConnectionPassive && !_flags.isListening){
        if(listen(_fd,SOMAXCONN) < 0)
            NSLog(@"VXConnection 0x%x: ERROR - listen() fails with errno %d\n",self, errno);
        else
            _flags.isListening = YES;
    }
}

- initLocalWithSocket:(int)fd address:(struct sockaddr_in*)addr
{
	@try{
		self = [self init];
		int reuseAddr = 1;
		unsigned addrlen = sizeof(struct sockaddr_in);
		if(addr->sin_family == AF_INET6)
			addrlen = sizeof(struct sockaddr_in6);
		_inetAddr = NSZoneMalloc([self zone], addrlen);
		memcpy((char*)_inetAddr, (char*)addr, addrlen);
		_fd = fd;
		if(setsockopt(_fd, SOL_SOCKET, SO_REUSEADDR, &reuseAddr, sizeof(reuseAddr)) < 0)
			NSLog(@"VXConnection 0x%x: failed to set socket option SO_REUSEADDR, errno = %d", self, errno);
		unsigned port = ntohs(addr->sin_port);
		if(bind(_fd, (struct sockaddr*)addr, addrlen) < 0)
			[NSException raise:ConnectionException
					format:@"Cannot bind socket to port %hu, 'bind' failed, errno %d", port, errno];
		if(port == 0){
			getsockname(_fd, (struct sockaddr*)_inetAddr, &addrlen);
			port = ntohs(((struct sockaddr_in*)_inetAddr)->sin_port);
		}
		if(enableLogging)
			NSLog(@"VXConnection 0x%x: bound socket %d to port %d\n", self, _fd, port);
		_flags.mode = VXConnectionPassive;
		setupSocket(_fd);
#if (TARGET_IPHONE_SIMULATOR || TARGET_OS_IPHONE)
		_hostname = [[[NSProcessInfo processInfo]hostName]copy];
#else
		_hostname = [[[NSHost currentHost]name]copy];
#endif
		unsigned flags = fcntl(_fd, F_GETFL, 0);
		flags |= O_NONBLOCK;
		fcntl(_fd, F_SETFL, flags);
		addConnection(self);
	}@catch(NSException *localException){
		NSLog(@"%@: %@\n", [localException name], [localException reason]);
		[self release];
		self = nil;
	}
	return self;
}

- initWithAcceptedSocket:(unsigned short)fd address:(struct sockaddr*)addr
{
	self = [self init];
    _fd = fd;
    _flags.mode = VXConnectionServer;
	switch(((struct sockaddr_in*)addr)->sin_family){
		case AF_INET:
			if(enableLogging)
				NSLog(@"VXConnection 0x%x: accepting IPv4 socket %d", self, _fd);
			_inetAddr = NSZoneMalloc([self zone], sizeof(struct sockaddr_in));
			memcpy(_inetAddr, addr, sizeof(struct sockaddr_in));
			break;
		case AF_INET6:
			if(enableLogging)
				NSLog(@"VXConnection 0x%x: accepting IPv6 socket %d", self, _fd);
			_inetAddr = NSZoneMalloc([self zone], sizeof(struct sockaddr_in6));
			memcpy(_inetAddr, addr, sizeof(struct sockaddr_in6));
			break;
	}
	
	setupSocket(_fd);
    _objects = [[NSMutableDictionary alloc]init];
	unsigned flags = fcntl(_fd, F_GETFL, 0);
	flags |= O_NONBLOCK;
	fcntl(_fd, F_SETFL, flags); // just for the case, flags must be inherited
    addConnection(self);
    [[NSNotificationCenter defaultCenter]postNotificationName:VXAcceptedConnection object:self];
    return self;
}

- initRemoteWithSocket:(int)fd host:(NSString*)hostname error:(NSError**)errObj
{
	self = [self init];
	_fd = fd;
	setupSocket(_fd);
	unsigned len = sizeof(struct sockaddr_in6);
	_inetAddr = NSZoneMalloc([self zone], len);
	memset((char *)_inetAddr, 0, len);
	if(getpeername(_fd, _inetAddr, &len) < 0){
		NSLog(@"VXConnection 0x%x: cannot get address for host '%@', getpeername failed, errno %d", 
			self, hostname, errno);
		if(errObj)
			*errObj = [NSError errorWithDomain:NSPOSIXErrorDomain code:errno
				userInfo:[NSDictionary dictionaryWithObjectsAndKeys:
					hostname, @"hostname", @"CORBACommunicationException", @"name", 
					@"getpeername failed", @"reason", nil]];
		[self release];
		return nil;
	}
	_flags.mode = VXConnectionClient;
	_objects =[[NSMutableDictionary alloc]init];
	addConnection(self);
	_hostname = [hostname copy];
	if(errObj)
		*errObj = nil;
	return self;
}

+ connectToHostNamed:(NSString*)hostName port:(int)port timeout:(NSTimeInterval)timeout error:(NSError**)errObj
{
    if([hostName length] == 0 || [hostName isEqualToString:@"localhost"])
		hostName = @"127.0.0.1";

	struct addrinfo hints, *r, *r0;
	int fd = -1;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	char portStr[8];
	sprintf(portStr, "%u", port);
	int error = getaddrinfo([hostName UTF8String], portStr, &hints, &r0);
	if(error){
		NSString* reason = [NSString stringWithFormat:@"Host '%@' is not reached, %s", hostName, gai_strerror(error)];
		NSLog(@"CORBACommunicationException: %@",  reason);
		if(errObj)
			*errObj = [NSError errorWithDomain:NSPOSIXErrorDomain code:EHOSTUNREACH 
				userInfo:[NSDictionary dictionaryWithObjectsAndKeys:reason, @"reason", hostName, @"hostname", nil]]; 
		return nil;
	}

	for(r = r0; r; r = r->ai_next){
		VXConnection* conn = lookupConnectionByAddr((struct sockaddr_in*)r->ai_addr);
		if(conn && [conn isValid]){
			freeaddrinfo(r0);
			*errObj = nil;
			return conn;
		}
	}
	for(r = r0; r; r = r->ai_next){
		if((fd = socket(r->ai_family, r->ai_socktype, r->ai_protocol)) < 0)
			continue;

		unsigned flags = fcntl(fd, F_GETFL, 0);
		fcntl(fd, F_SETFL, flags | O_NONBLOCK);
		if(connect(fd, r->ai_addr, r->ai_addrlen) < 0){
			if((error = errno) != EINPROGRESS){
				close(fd);
				fd = -1;
				continue;
			}
			NSDate* start = [NSDate date];
			double t = timeout;
			int found = 0;
			fd_set rset, wset;
			do{
				struct timeval tm = {floor(t), floor((t-floor(t))*1000000)};
				FD_ZERO(&rset);
				FD_SET(fd, &rset);
				wset = rset;
				if((found = select(fd + 1, &rset, &wset, NULL, t > 0? &tm : NULL)) < 0){
					if((error = errno) != EINTR)
						break;
					t -= [[NSDate date]timeIntervalSinceDate:start];
				}
			}while(found < 0 && t > 0.001);
			
			if(found > 0 && (FD_ISSET(fd, &rset) || FD_ISSET(fd, &wset))){
				unsigned len = sizeof(error);
				if(getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) == 0 && error == 0)
					break;
			}else if(found == 0)
				error = ETIMEDOUT;
			close(fd);
			fd = -1;
		}else
			break;
	}
	freeaddrinfo(r0);
	
	if(fd < 0){
		if(errObj)
			*errObj = [NSError errorWithDomain:NSPOSIXErrorDomain code:error 
				userInfo:[NSDictionary dictionaryWithObjectsAndKeys:
					hostName, @"hostname", [NSNumber numberWithInt:port], @"port", nil]];
		NSLog(@"CORBACommunicationException: cannot connect to host '%@:%u', errno = %d",  hostName, port, error);
		return nil;
	}
	return [[[self alloc] initRemoteWithSocket:fd host:hostName error:errObj]autorelease];
}

#define USE_IPv6 0

+ connection
{
    [_bindLock lock];
    @try{
		if(!_listener){
#if USE_IPv6
			struct sockaddr_in6 addr;
			memset((char*)&addr, 0, sizeof(addr));
			addr.sin6_family = AF_INET6;
			addr.sin6_port = htons(_listenerPort);
			int fd = socket(addr.sin6_family, SOCK_STREAM, IPPROTO_TCP);
#else
			struct sockaddr_in addr;
			memset((char*)&addr, 0, sizeof(addr));
			addr.sin_family = AF_INET;
			addr.sin_addr.s_addr = INADDR_ANY;
			addr.sin_port = htons(_listenerPort);
			int fd = socket(addr.sin_family, SOCK_STREAM, IPPROTO_TCP);
#endif
			if(fd < 0)
				[NSException raise:ConnectionException
					format:@"Cannot create socket, 'socket' failed, errno %d", errno];
			
			_listener = [[VXConnection alloc]initLocalWithSocket:fd address:(struct sockaddr_in*)&addr];
		}
    }@catch(NSException *localException){
		NSLog(@"VXConnection: Internal inconsistency: %@, %@\n", [localException name], [localException reason]);
    }
    [_bindLock unlock];
    return _listener;
}

- (int) fileDescriptor
{
    return _fd;
}

- (void) acceptConnection
{
    struct sockaddr_in6 remote;
    unsigned len = sizeof(struct sockaddr_in6);
    int fd = accept(_fd,(struct sockaddr*)&remote,&len);
    if(fd < 0){
		if(errno != EWOULDBLOCK && errno != ECONNABORTED && errno != EINTR)
			NSLog(@"VXConnection 0x%x: ERROR - unable to accept connection on socket %d, errno = %d\n", self, _fd, errno);
    }else{
        VXConnection* connection = [[VXConnection alloc]initWithAcceptedSocket:fd address:(struct sockaddr*)&remote];
        if(enableLogging)
            NSLog(@"VXConnection 0x%x: accepted socket %d\n", self, fd);
        [_connLock lock];
        if(connection)
            [_acceptedConnections addObject:[connection autorelease]];
        [_connLock unlock];
    }
}

#pragma mark ---- Message handling ----


+ (BOOL) handlesRequestsOnMainThread
{
    return handleInvocationsOnMainThread;
}

+ (void) setHandlesRequestsOnMainThread:(BOOL)flag
{
	handleInvocationsOnMainThread = flag? YES : NO;
}

- (int) mode
{
    return _flags.mode;
}

+ (void) processInvocation:(VXMessageHandler*)handler
{
    NSAutoreleasePool* mpool = [[NSAutoreleasePool alloc]init];
	[handler retain];
    do{
        NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
        @try{
			[handler handleMessage];
        }@catch(NSException *localException){
            NSLog(@"VXConnection: %@, %@\n",[localException name], [localException reason]);
            @try{
				NSError* error = [NSError errorWithDomain:CORBAErrorDomain code:kEInvocationException
					userInfo:[NSDictionary dictionaryWithObjectsAndKeys:
						[localException name], @"name", [localException reason], NSLocalizedDescriptionKey, nil]];
				[[handler connection] invalidateWithError:error];
			}@catch(NSException *localException){
                NSLog(@"VXConnection: %@, %@\n",[localException name], [localException reason]);
            }
        }
		[handler release];
		
		[_invocationLock lock];
		if([_invocationQueue count] > 0){
			handler = [[_invocationQueue objectAtIndex:0]retain];
			[_invocationQueue removeObjectAtIndex:0];
		}else
			handler = nil;
        [_invocationLock unlock];
        [pool release];
    }while(handler);
    [mpool release];
}

+ (void) processInvocationQueue
{
	[_invocationLock lock];
	if([_invocationQueue count] > 0){
		VXMessageHandler* handler = [[_invocationQueue objectAtIndex:0]retain];
		[_invocationQueue removeObjectAtIndex:0];
		[self processInvocation:handler];
	}
	[_invocationLock unlock];
}

+ (void) processInvocationInThread:(VXMessageHandler*)handler
{
	[self processInvocation:handler];
	[_invocationLock lock];
	numberOfThreads -= 1;
	[_invocationLock unlock];
	if(enableLogging)
		NSLog(@"Exiting invocation thread");
}

static void _queueInvocation(VXMessageHandler* handler)
{
	if([VXConnection handlesRequestsOnMainThread] || maxNumberOfThreads <= 0){
		if(enableLogging)
			NSLog(@"Queueing invocation %@ 0x%08x for main thread", handler, [handler requestId]);
		[VXConnection performSelectorOnMainThread:@selector(processInvocation:) withObject:handler waitUntilDone:NO];
	}else{
		if(enableLogging)
			NSLog(@"Queueing invocation %@ 0x%08x for detached thread", handler, [handler requestId]);
		[_invocationLock lock];
		if(numberOfThreads < maxNumberOfThreads){
			numberOfThreads += 1;
			// The current IDLDefinitionScope will be set for new thread in VXMessageHandler when needed
			[NSThread detachNewThreadSelector:@selector(processInvocationInThread:) 
					toTarget:[VXConnection class] withObject:handler];
		}else{
			[_invocationQueue addObject:handler];
		}
		[_invocationLock unlock];
	}
}

- (void) dispatchMessage:(VXMessage*)message
{
    @try{
		unsigned msgType = [message messageType];
		NSNumber* requestId = [NSNumber numberWithLong:[message requestId]];
		switch(msgType){
			case GIOP_ReplyType:
			case GIOP_LocateReplyType:{
				if(_trace)
					NSLog(@"RECEIVED REPLY for request <%x> from '%@:%u'\n", 
						[requestId unsignedIntValue], [self address], [[self port]unsignedShortValue]);
				[_lock lock];
				VXInvocation* inv = [[[_invocations objectForKey:requestId]retain]autorelease];
				if(inv){
					[_invocations removeObjectForKey:requestId];
					[inv setResponse:message];
				}else // No waiting invocation, ignore message
				if(enableLogging)
					NSLog(@"VXConnection:0x%x: ignoring reply '0x%08x' - no waiting invocation.\n", self, requestId);
				[_lock unlock];
			}
			break;
					
			case GIOP_CloseConnectionType:
				[self invalidateWithError:[NSError errorWithDomain:CORBAErrorDomain code:kECloseConnection userInfo:nil]];
				break;
			
			default:{
				VXMessageHandler* handler = [VXMessageHandler handlerWithMessage:message connection:self];
				_queueInvocation(handler);
				break;
			}
		}
    }@catch(NSException *localException){
		NSLog(@"%@: %@\n", [localException name], [localException reason]);
		//[localException raise];
    }
}

- (void) _readData
{
	if(!_readBuffer){
		_readBuffer = [[VXBuffer alloc]init];
		_flags.readState = VXReadHeader;
	}
	NSAssert(_readBuffer->length > 0, @"_readBuffer->length > 0");
	unsigned maxBufLength;
	unsigned optlen = sizeof(int);
	if(getsockopt(_fd, SOL_SOCKET, SO_RCVBUF, &maxBufLength, &optlen) < 0){
		maxBufLength = 65536;
		NSLog(@"VXConnection: getsockopt failed, errno %d", errno);
	}
	do{
		unsigned bufferLength = MIN(maxBufLength, _readBuffer->length);
		char* buffer = NSAllocateMemoryPages(bufferLength);
		if(buffer){
			int len = recv(_fd, buffer, bufferLength, 0);
			if(len < 0){
				NSDeallocateMemoryPages(buffer, bufferLength);
				if(errno != EAGAIN && errno != EINTR){
					NSError* error = [NSError errorWithDomain:NSPOSIXErrorDomain code:errno userInfo:nil];
					// close connection
					NSLog(@"VXConnection 0x%x: read error on socket %d, errno = %d", self, _fd, errno);
					[_readBuffer release];
					_readBuffer = nil;
					[self invalidateWithError:error];
				}
				return;
			}else if(len == 0){ // EOF
				if(enableLogging)
					NSLog(@"VXConnection 0x%x: received EOF, socket %d", self, _fd);
				NSDeallocateMemoryPages(buffer, bufferLength);
				// close connection
				if([_readBuffer->data length] > 0)
					NSLog(@"VXConnection 0x%x: CORBACommunicationException - read end of file, while expecting data", self);
				[_readBuffer release];
				_readBuffer = nil;
				NSError* error = [NSError errorWithDomain:CORBAErrorDomain code:kEEndOfFile userInfo:nil];
				[self invalidateWithError:error];
				return;
			}else{
				if(enableLogging)
					NSLog(@"VXConnection 0x%x: received %d bytes on socket %d, waiting for %d more", 
						self, len, _fd, _readBuffer->length - len);
				[_readBuffer->data appendBytes:buffer length:len];
				NSDeallocateMemoryPages(buffer, bufferLength);
				_readBuffer->length -= len;
				if(_readBuffer->length == 0){
					switch(_flags.readState){
						case VXReadHeader:
							_readBuffer->stream = [[IDLOctetStream alloc]initWithStreamData:_readBuffer->data];
							_readBuffer->header0 = [[VXMessageHeader alloc]initWithStream:_readBuffer->stream];
							_readBuffer->header = [_readBuffer->header0 retain];
							_readBuffer->length = [_readBuffer->header messageSize];
							if(_readBuffer->length > 0)
								_flags.readState = VXReadBody;
							else{
								VXMessage* msg = [[VXMessage alloc]initWithStream:_readBuffer->stream header:_readBuffer->header0];
								[self dispatchMessage:msg];
								[msg release];
								[_readBuffer release];
								_readBuffer = nil;
								_flags.readState = VXReadHeader;
								return;
							}
							break;
						
						case VXReadBody:
						case VXReadFragmentBody:
							if([_readBuffer->header moreFragments]){
								_flags.readState = VXReadFragmentHeader;
								_readBuffer->length = MSG_HEADER_SIZE;
							}else{
								VXMessage* msg = [[VXMessage alloc]initWithStream:_readBuffer->stream header:_readBuffer->header0];
								[self dispatchMessage:msg];
								[msg release];
								[_readBuffer release];
								_readBuffer = nil;
								_flags.readState = VXReadHeader;
								return;
							}
							break;
						
						case VXReadFragmentHeader:{
							NSRange r = {[_readBuffer->data length] - MSG_HEADER_SIZE, MSG_HEADER_SIZE};
							NSData* data = [_readBuffer->data subdataWithRange:r];
							IDLOctetStream* stream = [[IDLOctetStream alloc]initWithStreamData:(NSMutableData*)data];
							[_readBuffer->header release];
							_readBuffer->header = [[VXMessageHeader alloc]initWithStream:stream];
							[stream release];
							if([_readBuffer->header messageType] == GIOP_FragmentType){
								[_readBuffer->data setLength:[_readBuffer->data length] - MSG_HEADER_SIZE];
								_readBuffer->length = sizeof(int);
								_flags.readState = VXReadFragmentLength;
							}else if([_readBuffer->header messageType] == GIOP_CancelRequestType){
								NSLog(@"VXConnection 0x%x: received CancelRequest message type, fragment expected.", self);
								_readBuffer->length = sizeof(int);
								_flags.readState = VXReadFragmentCancel;
							}else{
								NSLog(@"VXConnection 0x%x: received incorrect message type %d, fragment expected.",
									self, [_readBuffer->header messageType]);
								[_readBuffer release];
								_readBuffer = nil;
								_flags.readState = VXReadHeader;
								NSError* error = [NSError errorWithDomain:CORBAErrorDomain code:kEInvalidMessage userInfo:nil];
								[self invalidateWithError:error];
								return;
							}
							break;
						}
						
						case VXReadFragmentLength:
							// skip fragment header
							_readBuffer->length = [_readBuffer->header messageSize] - sizeof(int);
							[_readBuffer->data setLength:[_readBuffer->data length] - sizeof(int)];
							_flags.readState = VXReadFragmentBody;
							break;
						
						case VXReadFragmentCancel:{
							VXMessage* msg = [[VXMessage alloc]initWithStream:_readBuffer->stream header:_readBuffer->header0];
							[self dispatchMessage:msg];
							[msg release];
							[_readBuffer release];
							_readBuffer = nil;
							_flags.readState = VXReadHeader;
							return;
						}
					}
				}
			}
		}
	}while(1);
}

- (void) _removeWriteBuffer
{
	[_lock lock];
	VXWriteBuffer* buf = _writeBuffer;
	_writeBuffer = [_writeBuffer->next retain];
	[buf release];
	[_lock unlock];
}

- (void) _writeData
{
	while(_writeBuffer){
		if(_writeBuffer->length > 0){
			int len = send(_fd, [_writeBuffer->data bytes] + _writeBuffer->offset, _writeBuffer->length, 0);
			if(len < 0){
				if(errno = EAGAIN)
					return;
				if(errno != EINTR){
					NSError* error = [NSError errorWithDomain:NSPOSIXErrorDomain code:errno userInfo:nil];
					NSLog(@"VXConnection 0x%x: write error, errno = %d", self, errno);
					[self invalidateWithError:error];
					return;
				}
			
			}else if(len > 0){
				_writeBuffer->offset += len;
				_writeBuffer->length -= len;
				if(_writeBuffer->length == 0) // remove buffer
					[self _removeWriteBuffer];
			} // len = 0 ?
		}else
			[self _removeWriteBuffer];
	}
}

static void _readConnections(NSSet* set)
{
    NSEnumerator* en = [set objectEnumerator];
    VXConnection* connection;
    while(connection = [en nextObject]){
		@try{
			switch([connection mode]){
				default:
					[connection _readData];
					break;
				case VXConnectionPassive:
					[connection acceptConnection];
					break;
			}
		}@catch(NSException *localException){
			NSLog(@"%@: %@", [localException name], [localException reason]);
		}
	}
}
static void _writeConnections(NSSet* set)
{
    NSEnumerator* en = [set objectEnumerator];
    VXConnection* connection;
    while(connection = [en nextObject]){
		@try{
			[connection _writeData];
		}@catch(NSException *localException){
			NSLog(@"%@: %@", [localException name], [localException reason]);
		}
	}
}

static NSSet* _raisedConnections(fd_set *fdSet)
{
	NSMutableSet* set = [NSMutableSet set];
	[_hlock lock];
#if USE_HASHTABLES
	NSHashEnumerator en = NSEnumerateHashTable(connections);
	VXConnection* connection;
	while(connection = NSNextHashEnumeratorItem(&en)){
		int fd = [connection fileDescriptor];
		if(fd >= 0 && FD_ISSET(fd, fdSet))
			[set addObject:connection];
	}
	NSEndHashTableEnumeration(&en);
#else
	NSEnumerator *en = [(NSSet*)connections objectEnumerator];
	VXConnection* connection;
	while(connection = [en nextObject]){
		int fd = [connection fileDescriptor];
		if(fd >= 0 && FD_ISSET(fd, fdSet))
			[set addObject:connection];
	}
#endif
	[_hlock unlock];
	return [set count] > 0? set : nil;
}

+ (void) _run:sender
{
    NSAutoreleasePool* threadPool = [[NSAutoreleasePool alloc]init];
    [_runLock lock];
    _running = YES;
    [_runLock unlock];
    [(VXConnection*)[self connection]listen];
    
    while(_running){
        NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
        
        @try{
			VXConnection* connection;
			unsigned short maxfd = 0;
			int found = 0;
			short fd;
			fd_set _read;
			fd_set _write;
			FD_ZERO(&_read);
			FD_ZERO(&_write);
			
			fd = [[_pipe fileHandleForReading]fileDescriptor];
			if(fd > maxfd)
				maxfd = fd;
			FD_SET(fd,&_read);
			[_hlock lock];
#if USE_HASHTABLES
			NSHashEnumerator en = NSEnumerateHashTable(connections);
			while(connection = NSNextHashEnumeratorItem(&en))
			{
				fd = [connection fileDescriptor];
				if(fd > maxfd)
					maxfd = fd;
				if(fd >= 0){
					FD_SET(fd,&_read);
					if(connection->_writeBuffer)
						FD_SET(fd, &_write);
				}
			}
			NSEndHashTableEnumeration(&en);
#else
			NSEnumerator *en = [(NSSet*)connections objectEnumerator];
			while(connection = [en nextObject])
			{
				fd = [connection fileDescriptor];
				if(fd > maxfd)
					maxfd = fd;
				if(fd >= 0){
					FD_SET(fd,&_read);
					if(connection->_writeBuffer)
						FD_SET(fd, &_write);
				}
			}
#endif
			[_hlock unlock];
			
			do{
				found = select(maxfd+1, &_read, &_write, NULL,[self _timeout]);
				if(found > 0){
					// Handle it
					NSSet* set = _raisedConnections(&_write);
					if(set)
						_writeConnections(set);
					
					set = _raisedConnections(&_read);
					if(set)
						//_handleRequests(set);
						_readConnections(set);

					fd = [[_pipe fileHandleForReading]fileDescriptor];
					if(FD_ISSET(fd,&_read)){
						[[_pipe fileHandleForReading]readDataOfLength:4];
						if(enableLogging)
							NSLog(@"VXConnection: resetting active connections\n");
					}
				}else if(found == 0){
					//[self fireTimer];
					break;
				}else{
					if(errno != EINTR && errno != EBADF){ // Interrupted or socket closed in other thread
						NSLog(@"VXConnection: Unrecoverable ERROR in 'select()', errno = %d. Exiting the run loop\n", errno);
						_running = NO;
						break;
					}else if(errno == EBADF)
						break;
				}
			}while(found <= 0);
			
        }@catch(NSException *localException){
			NSLog(@"%@, %@\n",[localException name],[localException reason]);
        }

        [pool release];
    }
    [_runLock lock];
    _running = NO;
    [_runLock unlock]; // finished
    [threadPool release];
}

+ (void) resetConnections
{
    unsigned char b[] = {0,0,0,0};
    [_runLock lock];
    if(_running){
        [[_pipe fileHandleForWriting]writeData:[NSData dataWithBytes:&b[0] length:4]];
        if(enableLogging)
            NSLog(@"VXConnection: notifying of updated connections\n");
    }
    [_runLock unlock];
}

+ (void) run
{
    [_runLock lock];
    if(!_running){
        [NSThread detachNewThreadSelector:@selector(_run:) toTarget:self withObject:nil];
        if(enableLogging)
            NSLog(@"VXConnection: detaching new thread.\n");
    }
    
    [_runLock unlock];
}

+ (BOOL) isRunning
{
    return _running;
}

- (void) sendMessage:message invocation:invocation
{
    NSException* exception = nil;
    if([self isValid]){
		NSNumber* requestId = [NSNumber numberWithLong:[invocation requestId]];
        @try{
			if(invocation){
				[_lock lock];
				[_invocations setObject:invocation forKey:requestId];
				[_lock unlock];
			}
			[message sendWithConnection:self];
        }@catch(NSException *localException){
			exception = localException;
			if(invocation){
				[_lock lock];
				[_invocations removeObjectForKey:requestId];
				[_lock unlock];
			}
        }

        if(exception)
            [exception raise];
    }else
        [NSException raise:CORBACommunicationException
            format:@"Connection is invalidated"];
}

- (void) sendMessage:message
{
    [self sendMessage:message invocation:nil];
}

- (void) invalidateWithError:(NSError*)error
{
    [_lock lock];
    if(!_flags.invalid){
        _flags.invalid = YES;
        close(_fd);
        _fd = -1;
		NSEnumerator* en = [_invocations objectEnumerator];
		VXInvocation* inv;
		while(inv = [en nextObject])
			[inv setResponse:nil];
		[_invocations removeAllObjects];
    }
    [_lock unlock];
    // Remove from hash, so it never will be returned
    [_hlock lock];
    if(!_flags.released){
#if USE_HASHTABLES
        NSHashRemove(connections,self);
#else
		CFSetRemoveValue(connections, self);
#endif
        _flags.released = YES;
    }
    [_hlock unlock];

	NSDictionary* info = error? [NSDictionary dictionaryWithObject:error forKey:@"error"] : nil;
    [[NSNotificationCenter defaultCenter]postNotificationName:VXInvalidateConnection object:self userInfo:info];
}
- (void) invalidate
{
	[self invalidateWithError:nil];
}

- (void) writeData:(NSData*)data
{
	if(_fd >= 0){
		VXWriteBuffer* wbuf = [[VXWriteBuffer alloc]initWithData:data];
		BOOL needsReset = NO;
		[_lock lock];
		VXWriteBuffer* last = _writeBuffer;
		if(last){
			while(last->next) last = last->next;
			last->next = wbuf;
		}else{
			_writeBuffer = wbuf;
			needsReset = YES;
		}
		[_lock unlock];
		if(needsReset)
			[VXConnection resetConnections];
	}else
        [NSException raise:CORBACommunicationException 
            format:@"Not connected"];
}

#pragma mark ---- vending objects ----

- objectForObjectID:(NSData*)objectKey
{
	id obj = nil;
	[_lock lock];
    if(objectKey){
        obj = [ORB objectForObjectID:objectKey];
        if(!obj)
            obj = [_objects objectForKey:objectKey];
        if(!obj){
            obj = [CORBAObject lookupObjectID:objectKey];
            if(obj)
                [_objects setObject:obj forKey:objectKey];
        }
    }
	[_lock unlock];
	return obj;
}

- (void) registerImplementationFor:(CORBAObject*)obj
{
	[_lock lock];
    NSData* objectID = [obj objectKey];
    if(!objectID){
        id impl = [obj implementation];
		objectID = [CORBAObject objectIdForObject:impl];
        [obj addProfileWithKey:objectID];
    }
    if(![_objects objectForKey:objectID])
        [_objects setObject:obj forKey:objectID];
	[_lock unlock];
}

- (void) removeObjectWithObjectID:(NSData*)key
{
	[_lock lock];
    [_objects removeObjectForKey:key];
	[_lock unlock];
}

+ (void) removeObject:(CORBAObject*)obj
{
    NSData* objectKey = [obj objectKey];
    if(objectKey){
        VXConnection* connection;
        [_hlock lock];
#if USE_HASHTABLES
        NSHashEnumerator en = NSEnumerateHashTable(connections);
        while(connection = NSNextHashEnumeratorItem(&en))
            [connection removeObjectWithObjectID:objectKey];
		NSEndHashTableEnumeration(&en);
#else
		NSEnumerator *en = [(NSSet*)connections objectEnumerator];
        while(connection = [en nextObject])
            [connection removeObjectWithObjectID:objectKey];
#endif
        [_hlock unlock];
    }
}

#pragma mark ---- Accessors ----

#define s6_addr32 __u6_addr.__u6_addr32

- (unsigned) hash 
{
	unsigned hash = 0;
	if(_inetAddr){
		struct sockaddr_in *addr = _inetAddr;
		switch(addr->sin_family){
			case AF_INET:
				hash = addr->sin_port ^ (addr->sin_family<<2) ^ addr->sin_addr.s_addr;
				break;
			case AF_INET6:{
				struct sockaddr_in6* addr6 = _inetAddr;
				int n;
				for(n=0;n<4;n++)
					hash = hash^addr6->sin6_addr.s6_addr32[n];
				hash ^= addr6->sin6_port ^ (addr6->sin6_family<<2);
			}
			break;
		}
	}
    return hash;
}

- (BOOL)isEqual:(VXConnection*)other
{
    if (!other || other->isa != isa || other->_inetAddr == NULL || _inetAddr == NULL) return NO;
	struct sockaddr_in *addr = _inetAddr;
	switch(addr->sin_family){
		case AF_INET:{
			struct sockaddr_in *otherAddr = other->_inetAddr;
			return otherAddr->sin_family == addr->sin_family &&
					otherAddr->sin_port == addr->sin_port &&
					otherAddr->sin_addr.s_addr == addr->sin_addr.s_addr;
		}
		break;
		case AF_INET6:{
			struct sockaddr_in6 *addr6 = _inetAddr;
			struct sockaddr_in6 *otherAddr6 = other->_inetAddr;
			return otherAddr6->sin6_family == addr6->sin6_family &&
					otherAddr6->sin6_port == addr6->sin6_port &&
					otherAddr6->sin6_addr.s6_addr32[0] == addr6->sin6_addr.s6_addr32[0] &&
					otherAddr6->sin6_addr.s6_addr32[1] == addr6->sin6_addr.s6_addr32[1] &&
					otherAddr6->sin6_addr.s6_addr32[2] == addr6->sin6_addr.s6_addr32[2] &&
					otherAddr6->sin6_addr.s6_addr32[3] == addr6->sin6_addr.s6_addr32[3];
		}
	}
	return NO;
}

- (BOOL) isValid
{
    return !_flags.invalid;
}

- (NSString*)remoteAddress
{
    if(_flags.mode != VXConnectionPassive)
        return [self address];
    else
        return nil;
}


- (NSString*)address
{
    if(!_address)
		_address = [addressHost((struct sockaddr_in *)_inetAddr) retain];
    return _address;
}


- (NSString*)remoteHostName
{
    if(_flags.mode != VXConnectionPassive)
        return [self hostName];
    else
        return nil;
}

- (NSString*)hostName
{
    if(!_hostname){
#if (TARGET_IPHONE_SIMULATOR || TARGET_OS_IPHONE)
		_hostname = [[self address]copy];
#else
        NSHost* host = [NSHost hostWithAddress:[self address]];
        NSString* hostname = [host name];
    	_hostname = hostname?hostname : [self address];
        [_hostname retain];
#endif
    }
    return _hostname;
}

- (NSNumber*)port
{
	struct sockaddr_in* addr = _inetAddr;
    return [NSNumber numberWithUnsignedShort:ntohs(addr->sin_port)];
}

- (NSNumber*)localPort
{
    if(_fd >= 0){
		struct sockaddr_in6 local;
		unsigned len = sizeof(struct sockaddr_in6);
		int ret = getsockname(_fd, (struct sockaddr *)&local, &len);
		if (ret < 0){
			NSLog(@"VXConnection 0x%x: socket %d has no name!\n", self, _fd);
			return nil;
		}
        return [NSNumber numberWithUnsignedShort:ntohs(local.sin6_port)];
    }else
        return nil;
}

- (NSNumber*) remotePort
{
    if(_fd >= 0){
        struct sockaddr_in6 remote;
        unsigned len = sizeof(struct sockaddr_in6);
        int ret = getpeername(_fd, (struct sockaddr *)&remote, &len);
        if (ret < 0){
            NSLog(@"VXConnection 0x%x: socket %d has no peer name! errno = %d\n", self, _fd, errno);
            return nil;
        }else{
            unsigned short port = ntohs(remote.sin6_port);
            return [NSNumber numberWithUnsignedShort:port];
        }
    }else
        return nil;
}

- (NSDictionary*) localPoint
{
    if(_fd >= 0){
		switch(_flags.mode){
			case VXConnectionPassive:
				return nil;
			
			case VXConnectionServer:
			case VXConnectionClient:{
				struct sockaddr_in6 mine6;
				unsigned len = sizeof(struct sockaddr_in6);
				int ret = getsockname(_fd, (struct sockaddr *)&mine6, &len);
				if (ret < 0){
					NSLog(@"VXConnection 0x%x: socket %d has no name! errno = %d\n", self, _fd, errno);
					return nil;
				}else{
					return [NSDictionary dictionaryWithObjectsAndKeys:
						addressPort((struct sockaddr_in*)&mine6), @"port",
						addressHost((struct sockaddr_in*)&mine6), @"host", 
						nil];
				}
			}
		}
    }
	return nil;
}

- (NSDictionary*) remotePoint
{
    if(_fd >= 0 && _flags.mode != VXConnectionPassive){
        struct sockaddr_in6 mine6;
        unsigned len = sizeof(struct sockaddr_in6);
        int ret = getpeername(_fd, (struct sockaddr *)&mine6, &len);
        if (ret < 0){
            NSLog(@"VXConnection 0x%x: socket %d has no peer name! errno = %d\n", self, _fd, errno);
            return nil;
        }else{
			return [NSDictionary dictionaryWithObjectsAndKeys:
				addressPort((struct sockaddr_in*)&mine6), @"port",
				addressHost((struct sockaddr_in*)&mine6), @"host", 
				nil];
		}
    }else
        return nil;
}

- localPointsContext
{
	if(_flags.mode == VXConnectionClient){
		NSString* host = [CORBAObject localHostAddress];
		NSNumber* port = [_listener port];
		if(host && port){
			NSDictionary* listenPoint = [NSDictionary dictionaryWithObjectsAndKeys:
				host, @"host", port, @"port", nil];
			NSDictionary* dict = [NSDictionary dictionaryWithObject:[NSArray arrayWithObject:listenPoint] forKey:@"listen_points"];
			IDLServiceContext* ctxt = [IDLServiceContext serviceContextWithContextId:ORB_BI_DIR_IIOP context:dict];
			return ctxt;
		}
	}
	return nil;
}

- (void) addListenPoints:(NSArray*)array
{
	if(_flags.mode == VXConnectionServer){
		[_connLock lock];
		if(!_listenPoints)
			_listenPoints = [[NSMutableDictionary alloc]init];
		NSEnumerator* en = [array objectEnumerator];
		NSDictionary* lp;
		while(lp = [en nextObject]){
			NSString* host = [lp valueForKey:@"host"];
			NSNumber* port = [lp valueForKey:@"port"];
			NSMutableSet* set = [_listenPoints objectForKey:host];
			if(set)
				[set addObject:port];
			else
				[_listenPoints setObject:[NSMutableSet setWithObject:port] forKey:host];
		}
		[_connLock unlock];
	}
}

- (BOOL) containsListenPointForHost:(NSString*)host port:(unsigned short)port
{
	BOOL does = NO;
	if(_listenPoints){
		[_connLock lock];
		NSSet* set = [_listenPoints objectForKey:host];
		if(set)
			does = [set containsObject:[NSNumber numberWithUnsignedShort:port]];
		[_connLock unlock];
	}
	return does;
}

+ findConnectionWithListenPointHost:(NSString*)host port:(unsigned short)port
{
	[_hlock lock];
	NSArray* array = [self allObjects];
	[_hlock unlock];
	NSEnumerator* en = [array objectEnumerator];
	VXConnection* conn;
	while(conn = [en nextObject])
		if([conn containsListenPointForHost:host port:port])
			return conn;
	return nil;
}

- (void) setStringEncodingsFromCodeSets:(IDLServiceContext*)codeSets
{
	if(codeSets && !_flags.stringEncodingSet){
		_stringEncoding = [IDLServiceContext stringEncodingWithCode:[[codeSets valueForKeyPath:@"context.char_data"]unsignedIntValue]];
		_wstringEncoding = [IDLServiceContext wstringEncodingWithCode:[[codeSets valueForKeyPath:@"context.wchar_data"]unsignedIntValue]];
		if(_stringEncoding == NSNotFound)
			_stringEncoding = NSUTF8StringEncoding;
		if(_wstringEncoding == NSNotFound)
			_wstringEncoding = NSUnicodeStringEncoding;
		_flags.stringEncodingSet = YES;
	}
}

- (void) setStringEncodings:(NSStringEncoding)charCode:(NSStringEncoding)wcharCode
{
	if(!_flags.stringEncodingSet){
		_stringEncoding = charCode;
		_wstringEncoding = wcharCode;
		_flags.stringEncodingSet = YES;
	}
}

- (BOOL) isStringEncodingsSet
{
	return _flags.stringEncodingSet;
}

- (NSStringEncoding) wstringEncoding
{
	return _wstringEncoding;
}

- (NSStringEncoding) stringEncoding
{
	return _stringEncoding;
}

- (BOOL) isCodeSetsSent
{
	return _flags.isCodeSetsSent;
}

- (void) setCodeSetsSent:(BOOL)flag
{
	_flags.isCodeSetsSent = flag? YES : NO;
}

- (struct sockaddr_in*) _inetAddr
{
	return (struct sockaddr_in*)_inetAddr;
}

- _initWithAddress:(struct sockaddr_in*)addr
{
	self = [super init];
	unsigned addrlen = sizeof(struct sockaddr_in);
	if(addr->sin_family == AF_INET6)
		addrlen = sizeof(struct sockaddr_in6);
	_inetAddr = NSZoneMalloc([self zone], addrlen);
	memcpy((char*)_inetAddr, (char*)addr, addrlen);
	return self;
}

@end

#pragma mark ---- Utilities ----

static void setupSocket(int fd)
{
	if(setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void*)&tcp_nodelay, sizeof(tcp_nodelay)) < 0)
            NSLog(@"VXConnection: setsockopt failed, errno %d", errno);
}

static VXConnection *lookupConnectionByAddr(struct sockaddr_in* inetAddr)
{
	if(enableLogging)
		NSLog(@"VXConnection: lookup connection %@(%@)", addressHost(inetAddr), addressPort(inetAddr));
    [_hlock lock];
#if USE_HASHTABLES
#if USE_DEFS
    struct { @defs(VXConnection) } shadow;
    shadow.isa = [VXConnection class];
    shadow._inetAddr = inetAddr;//FIXME
    VXConnection *conn = [(NSObject*)NSHashGet(connections, &shadow) retain];
#else
	VXConnection *shadow = [[VXConnection alloc]_initWithAddress:inetAddr];
    VXConnection *conn = [(NSObject*)NSHashGet(connections, shadow) retain];
#endif
#else
	VXConnection *shadow = [[VXConnection alloc]_initWithAddress:inetAddr];
	VXConnection *conn = [(NSObject*)CFSetGetValue(connections, shadow) retain];
#endif
    [_hlock unlock];
#if !USE_HASHTABLES || !USE_DEFS
	[shadow release];
#endif	
    return [conn autorelease];
}

static void addConnection(VXConnection* conn)
{
    [_hlock lock];
    @try{
#if USE_HASHTABLES
		NSHashInsertKnownAbsent(connections,conn);
#else
		CFSetAddValue(connections, conn);
#endif  
		[VXConnection resetConnections];
		[VXConnection run];
		if(enableLogging)
			NSLog(@"VXConnection: added connection %@(%@)", addressHost([conn _inetAddr]), addressPort([conn _inetAddr]));
    }@catch(NSException *localException){
		NSLog(@"VXConnection: Internal inconsistency: %@, %@\n", [localException name], [localException reason]);
    }
    [_hlock unlock];
}

static NSString* addressHost(struct sockaddr_in* addr)
{
	char str[INET6_ADDRSTRLEN];
	const char* p = NULL;
	switch(addr->sin_family){
		case AF_INET:{
			if(p = inet_ntop(AF_INET, &addr->sin_addr, &str[0], INET_ADDRSTRLEN))
				return [[[NSString alloc]initWithBytes:p 
						length:strlen(p) encoding:NSASCIIStringEncoding]autorelease];
			break;
		}
		case AF_INET6:{
			struct sockaddr_in6 *addr6 = (struct sockaddr_in6*)addr;
			if(p = inet_ntop(AF_INET6, &addr6->sin6_addr, &str[0], INET6_ADDRSTRLEN))
				return [[[NSString alloc]initWithBytes:p 
						length:strlen(p) encoding:NSASCIIStringEncoding]autorelease];
			break;
		}
	}
	return nil;
}

static NSNumber* addressPort(struct sockaddr_in* addr)
{
	return [NSNumber numberWithUnsignedInt:ntohs(addr->sin_port)];
}

@implementation VXBuffer

- (void) dealloc
{
	[data release];
	[header release];
	[header release];
	[stream release];
	[super dealloc];
}

- init
{
	self = [super init];
	data = [[NSMutableData alloc]init];
	length = MSG_HEADER_SIZE;
	return self;
}

@end

@implementation VXWriteBuffer

- (void) dealloc
{
	[data release];
	[next release];
	[super dealloc];
}

- initWithData:(NSData*)aData
{
	self = [super init];
	data = [aData retain];
	length = [data length];
	return self;
}

@end
